"""Content Class for mongo backend."""

from datetime import datetime
from typing import Any, Optional

from bson import ObjectId

from forum.backends.mongodb.contents import BaseContents
from forum.backends.mongodb.users import Users
from forum.utils import get_handler_by_name


class CommentThread(BaseContents):
    """
    CommentThread class for cs_comments_service content model
    """

    index_name = "comment_threads"
    content_type = "CommentThread"

    def delete(self, _id: str) -> int:
        """Delete CommentThread"""
        result = super().delete(_id)
        get_handler_by_name("comment_thread_deleted").send(
            sender=self.__class__, comment_thread_id=_id
        )
        Users().delete_read_state_by_thread_id(_id)
        return result

    @classmethod
    def mapping(cls) -> dict[str, Any]:
        """
        Mapping function for the Thread class
        """
        return {
            "dynamic": "false",
            "properties": {
                "title": {
                    "type": "text",
                    "boost": 5.0,
                    "store": True,
                    "term_vector": "with_positions_offsets",
                },
                "body": {
                    "type": "text",
                    "store": True,
                    "term_vector": "with_positions_offsets",
                },
                "created_at": {"type": "date"},
                "updated_at": {"type": "date"},
                "last_activity_at": {"type": "date"},
                "comment_count": {"type": "integer"},
                "votes_point": {"type": "integer"},
                "context": {"type": "keyword"},
                "course_id": {"type": "keyword"},
                "commentable_id": {"type": "keyword"},
                "author_id": {"type": "keyword"},
                "group_id": {"type": "integer"},
                "id": {"type": "keyword"},
                "thread_id": {"type": "keyword"},
            },
        }

    @classmethod
    def doc_to_hash(cls, doc: dict[str, Any]) -> dict[str, Any]:
        """
        Converts thread document to the dict
        """
        return {
            "id": str(doc.get("_id")),
            "title": doc.get("title"),
            "body": doc.get("body"),
            "created_at": doc.get("created_at"),
            "updated_at": doc.get("updated_at"),
            "last_activity_at": doc.get("last_activity_at"),
            "comment_count": doc.get("comment_count"),
            "votes_point": doc.get("votes", {}).get("point"),
            "context": doc.get("context"),
            "course_id": doc.get("course_id"),
            "commentable_id": doc.get("commentable_id"),
            "author_id": doc.get("author_id"),
            "group_id": doc.get("group_id"),
            "thread_id": str(doc.get("_id")),
        }

    def insert(
        self,
        title: str,
        body: str,
        course_id: str,
        commentable_id: str,
        author_id: str,
        author_username: Optional[str] = None,
        anonymous: bool = False,
        anonymous_to_peers: bool = False,
        thread_type: str = "discussion",
        context: str = "course",
        pinned: bool = False,
        visible: bool = True,
        abuse_flaggers: Optional[list[str]] = None,
        historical_abuse_flaggers: Optional[list[str]] = None,
        group_id: Optional[int] = None,
    ) -> str:
        """
        Inserts a new thread document into the database.

        Args:
            title (str): The title of the thread.
            body (str): The body content of the thread.
            course_id (str): The ID of the course the thread is associated with.
            commentable_id (str): The ID of the commentable entity.
            author_id (str): The ID of the author who created the thread.
            author_username (str): The username of the author.
            anonymous (bool, optional): Whether the thread is posted anonymously. Defaults to False.
            anonymous_to_peers (bool, optional): Whether the thread is anonymous to peers. Defaults to False.
            thread_type (str, optional): The type of the thread, either 'question' or 'discussion'.
            context (str, optional): The context of the thread, either 'course' or 'standalone'.
            pinned (bool): Whether the thread is pinned. Defaults to False.
            visible (bool): Whether the thread is visible. Defaults to True.
            abuse_flaggers: A list of users who flagged the thread for abuse.
            historical_abuse_flaggers: A list of users who historically flagged the thread for abuse.

        Raises:
            ValueError: If `thread_type` is not 'question' or 'discussion'.
            ValueError: If `context` is not 'course' or 'standalone'.

        Returns:
            str: The ID of the inserted document.
        """
        if thread_type not in ["question", "discussion"]:
            raise ValueError("Invalid thread_type")

        if context not in ["course", "standalone"]:
            raise ValueError("Invalid context")

        if abuse_flaggers is None:
            abuse_flaggers = []
        if historical_abuse_flaggers is None:
            historical_abuse_flaggers = []

        date = datetime.now()
        thread_data = {
            "votes": self.get_votes_dict(up=[], down=[]),
            "thread_type": thread_type,
            "context": context,
            "comment_count": 0,
            "at_position_list": [],
            "title": title,
            "body": body,
            "course_id": course_id,
            "commentable_id": commentable_id,
            "_type": self.content_type,
            "anonymous": anonymous,
            "anonymous_to_peers": anonymous_to_peers,
            "closed": False,
            "author_id": author_id,
            "author_username": author_username or self.get_author_username(author_id),
            "created_at": date,
            "updated_at": date,
            "last_activity_at": date,
            "pinned": pinned,
            "visible": visible,
            "abuse_flaggers": abuse_flaggers,
            "historical_abuse_flaggers": historical_abuse_flaggers,
        }
        if group_id:
            thread_data["group_id"] = group_id

        result = self._collection.insert_one(thread_data)
        thread_id = str(result.inserted_id)

        # Notify Thread inserted
        get_handler_by_name("comment_thread_inserted").send(
            sender=self.__class__, comment_thread_id=thread_id
        )
        return thread_id

    def update(
        self,
        thread_id: str,
        thread_type: Optional[str] = None,
        title: Optional[str] = None,
        body: Optional[str] = None,
        course_id: Optional[str] = None,
        anonymous: Optional[bool] = None,
        anonymous_to_peers: Optional[bool] = None,
        commentable_id: Optional[str] = None,
        at_position_list: Optional[list[str]] = None,
        closed: Optional[bool] = None,
        context: Optional[str] = None,
        author_id: Optional[str] = None,
        author_username: Optional[str] = None,
        votes: Optional[dict[str, int]] = None,
        abuse_flaggers: Optional[list[str]] = None,
        historical_abuse_flaggers: Optional[list[str]] = None,
        closed_by: Optional[str] = None,
        pinned: Optional[bool] = None,
        comments_count: Optional[int] = None,
        endorsed: Optional[bool] = None,
        edit_history: Optional[list[dict[str, Any]]] = None,
        original_body: Optional[str] = None,
        editing_user_id: Optional[str] = None,
        edit_reason_code: Optional[str] = None,
        close_reason_code: Optional[str] = None,
        closed_by_id: Optional[str] = None,
        group_id: Optional[int] = None,
        skip_timestamp_update: bool = False,
    ) -> int:
        """
        Updates a thread document in the database.

        Args:
            thread_id: ID of thread to update.
            thread_type: The type of the thread, either 'question' or 'discussion'.
            title: The title of the thread.
            body: The body content of the thread.
            course_id: The ID of the course the thread is associated with.
            anonymous: Whether the thread is posted anonymously.
            anonymous_to_peers: Whether the thread is anonymous to peers.
            commentable_id: The ID of the commentable entity.
            at_position_list: A list of positions for @mentions.
            closed: Whether the thread is closed.
            context: The context of the thread, either 'course' or 'standalone'.
            author_id: The ID of the author who created the thread.
            author_username: The username of the author.
            votes: The votes for the thread.
            abuse_flaggers: A list of users who flagged the thread for abuse.
            historical_abuse_flaggers: A list of users who historically flagged the thread for abuse.
            closed_by: The ID of the user who closed the thread.
            pinned: Whether the thread is pinned.
            comments_count: The number of comments on the thread.
            endorsed: Whether the thread is endorsed.
            skip_timestamp_update: Whether to skip updating the timestamp (default: False).

        Returns:
            int: The number of documents modified.
        """
        fields = [
            ("thread_type", thread_type),
            ("title", title),
            ("body", body),
            ("course_id", course_id),
            ("anonymous", anonymous),
            ("anonymous_to_peers", anonymous_to_peers),
            ("commentable_id", commentable_id),
            ("at_position_list", at_position_list),
            ("closed", closed),
            ("context", context),
            ("author_id", author_id),
            ("author_username", author_username),
            ("votes", votes),
            ("abuse_flaggers", abuse_flaggers),
            ("historical_abuse_flaggers", historical_abuse_flaggers),
            ("closed_by", closed_by),
            ("pinned", pinned),
            ("comment_count", comments_count),
            ("endorsed", endorsed),
            ("close_reason_code", close_reason_code),
            ("closed_by_id", closed_by_id),
            ("group_id", group_id),
        ]
        update_data: dict[str, Any] = {
            field: value for field, value in fields if value is not None
        }
        if not closed and (close_reason_code and closed_by_id):
            update_data["closed_by_id"] = None
            update_data["close_reason_code"] = None

        if editing_user_id:
            edit_history = [] if edit_history is None else edit_history
            edit_history.append(
                {
                    "author_id": editing_user_id,
                    "original_body": original_body,
                    "reason_code": edit_reason_code,
                    "editor_username": self.get_author_username(editing_user_id),
                    "created_at": datetime.now(),
                }
            )
            update_data["edit_history"] = edit_history

        if not skip_timestamp_update:
            date = datetime.now()
            update_data["updated_at"] = date
        result = self._collection.update_one(
            {"_id": ObjectId(thread_id)},
            {"$set": update_data},
        )

        # Notify thread updated
        get_handler_by_name("comment_thread_updated").send(
            sender=self.__class__, comment_thread_id=thread_id
        )
        return result.modified_count

    def get_author_username(self, author_id: str) -> str | None:
        """Return username for the respective author_id(user_id)"""
        user = Users().get(author_id)
        return user.get("username") if user else None



"""Forum backend interface class."""

from typing import Any, Optional


class AbstractBackend:
    """Abstract backend interface class."""

    @classmethod
    def update_stats_for_course(
        cls, user_id: str, course_id: str, **kwargs: Any
    ) -> None:
        """Update statistics for a course."""
        raise NotImplementedError

    @classmethod
    def flag_as_abuse(
        cls, user_id: str, entity_id: str, **kwargs: Any
    ) -> dict[str, Any]:
        """Flag an entity as abuse."""
        raise NotImplementedError

    @classmethod
    def update_stats_after_unflag(
        cls, user_id: str, entity_id: str, has_no_historical_flags: bool, **kwargs: Any
    ) -> None:
        """Update statistics after unflagging an entity."""
        raise NotImplementedError

    @classmethod
    def un_flag_as_abuse(
        cls, user_id: str, entity_id: str, **kwargs: Any
    ) -> dict[str, Any]:
        """Unflag an entity as abuse."""
        raise NotImplementedError

    @classmethod
    def un_flag_all_as_abuse(cls, entity_id: str, **kwargs: Any) -> dict[str, Any]:
        """Unflag all entities as abuse."""
        raise NotImplementedError

    @staticmethod
    def update_vote(
        content_id: str,
        user_id: str,
        vote_type: str = "",
        is_deleted: bool = False,
        **kwargs: Any
    ) -> bool:
        """Update vote for a content."""
        raise NotImplementedError

    @classmethod
    def upvote_content(cls, entity_id: str, user_id: str, **kwargs: Any) -> bool:
        """Upvote a content."""
        raise NotImplementedError

    @classmethod
    def downvote_content(cls, entity_id: str, user_id: str, **kwargs: Any) -> bool:
        """Downvote a content."""
        raise NotImplementedError

    @classmethod
    def remove_vote(cls, entity_id: str, user_id: str, **kwargs: Any) -> bool:
        """Remove a vote for a content."""
        raise NotImplementedError

    @staticmethod
    def validate_thread_and_user(
        user_id: str, thread_id: str
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Validate a thread and user."""
        raise NotImplementedError

    @staticmethod
    def pin_unpin_thread(thread_id: str, action: str) -> None:
        """Pin or unpin a thread."""
        raise NotImplementedError

    @staticmethod
    def get_pinned_unpinned_thread_serialized_data(
        user_id: str, thread_id: str, serializer_class: Any
    ) -> dict[str, Any]:
        """Get pinned or unpinned thread serialized data."""
        raise NotImplementedError

    @classmethod
    def handle_pin_unpin_thread_request(
        cls, user_id: str, thread_id: str, action: str, serializer_class: Any
    ) -> dict[str, Any]:
        """Handle pin or unpin thread request."""
        raise NotImplementedError

    @staticmethod
    def get_abuse_flagged_count(thread_ids: list[str]) -> dict[str, int]:
        """Get abuse flagged count."""
        raise NotImplementedError

    @staticmethod
    def get_read_states(
        thread_ids: list[str], user_id: str, course_id: str
    ) -> dict[str, list[Any]]:
        """Get read states."""
        raise NotImplementedError

    @staticmethod
    def get_endorsed(thread_ids: list[str]) -> dict[str, bool]:
        """Get endorsed."""
        raise NotImplementedError

    @staticmethod
    def get_user_read_state_by_course_id(
        user_id: str, course_id: str
    ) -> dict[str, Any]:
        """Get user read state by course id."""
        raise NotImplementedError

    @classmethod
    def handle_threads_query(
        cls,
        comment_thread_ids: list[str],
        user_id: str,
        course_id: str,
        group_ids: list[int],
        author_id: Optional[str],
        thread_type: Optional[str],
        filter_flagged: bool,
        filter_unread: bool,
        filter_unanswered: bool,
        filter_unresponded: bool,
        count_flagged: bool,
        sort_key: str,
        page: int,
        per_page: int,
        context: str = "course",
        raw_query: bool = False,
        commentable_ids: Optional[list[str]] = None,
        is_moderator: bool = False,
    ) -> dict[str, Any]:
        """Handle threads query."""
        raise NotImplementedError

    @staticmethod
    def prepare_thread(
        thread_id: str,
        is_read: bool,
        unread_count: int,
        is_endorsed: bool,
        abuse_flagged_count: int,
    ) -> dict[str, Any]:
        """Prepare thread."""
        raise NotImplementedError

    @classmethod
    def threads_presentor(
        cls,
        thread_ids: list[str],
        user_id: str,
        course_id: str,
        count_flagged: bool = False,
    ) -> list[dict[str, Any]]:
        """Threads presenter."""
        raise NotImplementedError

    @staticmethod
    def get_username_from_id(user_id: str) -> Optional[str]:
        """Get username from id."""
        raise NotImplementedError

    @staticmethod
    def validate_object(model: str, obj_id: str) -> Any:
        """Validate object."""
        raise NotImplementedError

    @staticmethod
    def find_subscribed_threads(
        user_id: str, course_id: Optional[str] = None
    ) -> list[str]:
        """Find subscribed threads."""
        raise NotImplementedError

    @staticmethod
    def subscribe_user(
        user_id: str, source_id: str, source_type: str
    ) -> dict[str, Any] | None:
        """Subscribe user."""
        raise NotImplementedError

    @staticmethod
    def unsubscribe_user(user_id: str, source_id: str, source_type: str) -> None:
        """Unsubscribe user."""
        raise NotImplementedError

    @staticmethod
    def delete_comments_of_a_thread(thread_id: str) -> None:
        """Delete comments of a thread."""
        raise NotImplementedError

    @staticmethod
    def delete_subscriptions_of_a_thread(thread_id: str) -> None:
        """Delete subscriptions of a thread."""
        raise NotImplementedError

    @staticmethod
    def validate_params(params: dict[str, Any], user_id: Optional[str] = None) -> Any:
        """Validate params."""
        raise NotImplementedError

    @classmethod
    def get_threads(
        cls,
        params: dict[str, Any],
        user_id: str,
        serializer: Any,
        thread_ids: list[str],
    ) -> dict[str, Any]:
        """Get threads."""
        raise NotImplementedError

    @staticmethod
    def get_commentables_counts_based_on_type(course_id: str) -> dict[str, Any]:
        """Get commentables counts based on type."""
        raise NotImplementedError

    @classmethod
    def get_user_voted_ids(cls, user_id: str, vote: str) -> list[str]:
        """Get user voted ids."""
        raise NotImplementedError

    @staticmethod
    def filter_standalone_threads(comment_ids: list[str]) -> list[str]:
        """Filter standalone threads."""
        raise NotImplementedError

    @classmethod
    def user_to_hash(
        cls, user_id: str, params: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """User to hash."""
        raise NotImplementedError

    @staticmethod
    def replace_username_in_all_content(user_id: str, username: str) -> None:
        """Replace username in all content."""
        raise NotImplementedError

    @staticmethod
    def unsubscribe_all(user_id: str) -> None:
        """Unsubscribe all."""
        raise NotImplementedError

    @staticmethod
    def retire_all_content(user_id: str, username: str) -> None:
        """Retire all content."""
        raise NotImplementedError

    @staticmethod
    def find_or_create_read_state(user_id: str, thread_id: str) -> dict[str, Any]:
        """Find or create read state."""
        raise NotImplementedError

    @classmethod
    def mark_as_read(cls, user_id: str, thread_id: str) -> None:
        """Mark as read."""
        raise NotImplementedError

    @staticmethod
    def find_or_create_user_stats(user_id: str, course_id: str) -> dict[str, Any]:
        """Find or create user stats."""
        raise NotImplementedError

    @staticmethod
    def update_user_stats_for_course(user_id: str, stat: dict[str, Any]) -> None:
        """Update user stats for course."""
        raise NotImplementedError

    @classmethod
    def build_course_stats(cls, author_id: str, course_id: str) -> None:
        """Build course stats."""
        raise NotImplementedError

    @classmethod
    def update_all_users_in_course(cls, course_id: str) -> list[str]:
        """Update all users in course."""
        raise NotImplementedError

    @staticmethod
    def get_user_by_username(username: str | None) -> dict[str, Any] | None:
        """Get user by username."""
        raise NotImplementedError

    @staticmethod
    def find_or_create_user(
        user_id: str, username: Optional[str] = "", default_sort_key: Optional[str] = ""
    ) -> str:
        """Find or create user."""
        raise NotImplementedError

    @staticmethod
    def get_comment(comment_id: str) -> dict[str, Any] | None:
        """Get comment."""
        raise NotImplementedError

    @staticmethod
    def get_thread(thread_id: str) -> dict[str, Any] | None:
        """Get thread."""
        raise NotImplementedError

    @staticmethod
    def get_comments(**kwargs: Any) -> list[dict[str, Any]]:
        """Get comments."""
        raise NotImplementedError

    @staticmethod
    def get_comments_count(**kwargs: Any) -> int:
        """Get comments count."""
        raise NotImplementedError

    @classmethod
    def create_comment(cls, data: dict[str, Any]) -> Any:
        """Create comment."""
        raise NotImplementedError

    @staticmethod
    def delete_comment(comment_id: str) -> None:
        """Delete comment."""
        raise NotImplementedError

    @staticmethod
    def update_comment(comment_id: str, **kwargs: Any) -> int:
        """Update comment."""
        raise NotImplementedError

    @staticmethod
    def get_thread_id_from_comment(comment_id: str) -> dict[str, Any] | None:
        """Get thread id from comment."""
        raise NotImplementedError

    @staticmethod
    def get_user(user_id: str, get_full_dict: bool = True) -> dict[str, Any] | None:
        """Get user."""
        raise NotImplementedError

    @staticmethod
    def get_subscription(
        subscriber_id: str, source_id: str, **kwargs: Any
    ) -> dict[str, Any] | None:
        """Get subscription."""
        raise NotImplementedError

    @staticmethod
    def get_subscriptions(query: dict[str, Any]) -> list[dict[str, Any]]:
        """Get subscriptions."""
        raise NotImplementedError

    @staticmethod
    def delete_thread(thread_id: str) -> int:
        """Delete thread."""
        raise NotImplementedError

    @staticmethod
    def create_thread(data: dict[str, Any]) -> str:
        """Create thread."""
        raise NotImplementedError

    @staticmethod
    def update_thread(thread_id: str, **kwargs: Any) -> int:
        """Update thread."""
        raise NotImplementedError

    @staticmethod
    def get_filtered_threads(
        query: dict[str, Any], ids_only: bool = False
    ) -> list[dict[str, Any]]:
        """Get filtered threads."""
        raise NotImplementedError

    @staticmethod
    def update_user(user_id: str, data: dict[str, Any]) -> int:
        """Update user."""
        raise NotImplementedError

    @staticmethod
    def get_thread_id_by_comment_id(parent_comment_id: str) -> str:
        """
        The thread Id from the parent comment.
        """
        raise NotImplementedError

    @staticmethod
    def update_comment_and_get_updated_comment(
        comment_id: str,
        body: Optional[str] = None,
        course_id: Optional[str] = None,
        user_id: Optional[str] = None,
        anonymous: Optional[bool] = False,
        anonymous_to_peers: Optional[bool] = False,
        endorsed: Optional[bool] = False,
        closed: Optional[bool] = False,
        editing_user_id: Optional[str] = None,
        edit_reason_code: Optional[str] = None,
        endorsement_user_id: Optional[str] = None,
    ) -> dict[str, Any] | None:
        """Update comment and get updated comment."""
        raise NotImplementedError

    @staticmethod
    def get_contents(**kwargs: Any) -> list[dict[str, Any]]:
        """Get contents."""
        raise NotImplementedError

    @staticmethod
    def get_users(**kwargs: Any) -> list[dict[str, Any]]:
        """Get users."""
        raise NotImplementedError

    @staticmethod
    def get_user_sort_criterion(sort_by: str) -> dict[str, Any]:
        """Get sort criterion."""
        raise NotImplementedError

    @staticmethod
    def get_thread_index_name() -> str:
        """Get the name of the thread index."""
        return "comment_threads"

    @staticmethod
    def get_votes_dict(up: list[str], down: list[str]) -> dict[str, Any]:
        """
        Calculates and returns the vote summary for a thread.

        Args:
            up (list): A list of user IDs who upvoted the thread.
            down (list): A list of user IDs who downvoted the thread.

        Returns:
            dict: A dictionary containing the vote summary with the following keys:
                - "up" (list): The list of user IDs who upvoted.
                - "down" (list): The list of user IDs who downvoted.
                - "up_count" (int): The count of upvotes.
                - "down_count" (int): The count of downvotes.
                - "count" (int): The total number of votes (upvotes + downvotes).
                - "point" (int): The vote score (upvotes - downvotes).
        """
        up = up or []
        down = down or []
        votes = {
            "up": up,
            "down": down,
            "up_count": len(up),
            "down_count": len(down),
            "count": len(up) + len(down),
            "point": len(up) - len(down),
        }
        return votes

    @staticmethod
    def find_thread(**kwargs: Any) -> Optional[dict[str, Any]]:
        """
        Retrieves a first matching thread from the database.
        """
        raise NotImplementedError

    @staticmethod
    def find_comment(
        is_parent_comment: bool = True, with_abuse_flaggers: bool = False, **kwargs: Any
    ) -> Optional[dict[str, Any]]:
        """
        Retrieves a first matching comment from the database.
        """
        raise NotImplementedError

    @staticmethod
    def get_user_contents_by_username(username: str) -> list[dict[str, Any]]:
        """
        Retrieve all threads and comments authored by a specific user.
        """
        raise NotImplementedError



"""
Serializer for the comment data.
"""

from typing import Any

from rest_framework import serializers

from forum.serializers.contents import ContentSerializer
from forum.serializers.custom_datetime import CustomDateTimeField
from forum.utils import prepare_comment_data_for_get_children


class EndorsementSerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for handling endorsement of a comment

    Attributes:
        user_id (str): The endorsement user id.
        time (datetime): The timestamp of when the user has endorsed the comment.
    """

    user_id = serializers.CharField()
    time = CustomDateTimeField()

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError


class CommentSerializer(ContentSerializer):
    """
    Serializer for handling user comments on threads.

    Inherits from ContentSerializer.

    Attributes:
        endorsed (bool): Whether the comment is endorsed by an authority.
        depth (int): The depth of the comment in a nested comment structure.
        thread_id (str): The ID of the thread the comment belongs to.
        parent_id (str or None): The ID of the parent comment, if any.
        child_count (int): The number of child comments nested under this comment.
        sk (str or None): sk field, has ids data in it.
        endorsement: saves endorsement data.
    """

    endorsed = serializers.BooleanField(default=False)
    depth = serializers.IntegerField(default=0)
    thread_id = serializers.CharField(source="comment_thread_id")
    parent_id = serializers.CharField(default=None, allow_null=True)
    child_count = serializers.IntegerField(default=0)
    sk = serializers.CharField(default=None, required=False, allow_null=True)
    endorsement = EndorsementSerializer(default=None, required=False, allow_null=True)
    children = serializers.SerializerMethodField()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        exclude_fields = kwargs.pop("exclude_fields", None)
        self.backend = kwargs.pop("backend")
        super().__init__(*args, **kwargs)
        if exclude_fields:
            for field in exclude_fields:
                self.fields.pop(field, None)

    def get_children(self, obj: Any) -> list[dict[str, Any]]:
        """Get comments of a thread."""
        if not self.context.get("recursive", False):
            return []

        children = self.backend.get_comments(
            parent_id=obj["_id"],
            depth=1,
            sort=self.context.get("sort", -1),
        )
        children_data = prepare_comment_data_for_get_children(children)
        serializer = CommentSerializer(
            children_data,
            many=True,
            context={"recursive": False},
            exclude_fields=["sk"],
            backend=self.backend,
        )
        return list(serializer.data)

    def to_representation(self, instance: Any) -> dict[str, Any]:
        comment = super().to_representation(instance)
        comment.pop("historical_abuse_flaggers")
        if comment["parent_id"] == "None":
            comment["parent_id"] = None

        thread = self.backend.get_thread(comment["thread_id"])
        comment_from_db = self.backend.get_comment(comment["id"])
        if (
            not comment["endorsed"]
            and comment_from_db
            and "endorsement" not in comment_from_db
            and thread
            and thread["thread_type"] == "question"
        ):
            comment.pop("endorsement", None)
        return comment

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError



"""Model util function for db operations."""

import math
from datetime import datetime, timezone
from typing import Any, Optional

from bson import ObjectId, errors as bson_errors
from django.core.exceptions import ObjectDoesNotExist

from forum.backends.backend import AbstractBackend
from forum.backends.mongodb import (
    Comment,
    CommentThread,
    Contents,
    Subscriptions,
    Users,
)
from forum.constants import RETIRED_BODY, RETIRED_TITLE
from forum.utils import (
    ForumV2RequestError,
    get_group_ids_from_params,
    get_sort_criteria,
    make_aware,
    str_to_bool,
)


class MongoBackend(AbstractBackend):
    """Mongodb Backend API."""

    @classmethod
    def update_stats_for_course(
        cls, user_id: str, course_id: str, **kwargs: Any
    ) -> None:
        """Update stats for a course."""
        user = Users().get(user_id)
        if not user:
            return
        course_stats = user.get("course_stats", [])
        for course_stat in course_stats:
            if course_stat["course_id"] == course_id:
                course_stat.update(
                    {
                        k: course_stat[k] + v
                        for k, v in kwargs.items()
                        if k in course_stat
                    }
                )
                Users().update(
                    user_id,
                    course_stats=course_stats,
                )
                return
        cls.build_course_stats(user["_id"], course_id)

    @classmethod
    def flag_as_abuse(
        cls, user_id: str, entity_id: str, **kwargs: Any
    ) -> dict[str, Any]:
        """
        Flag an entity as abuse.

        Args:
            user (dict[str, Any]): The user who is flagging the entity as abuse.
            entity (dict[str, Any]): The entity being flagged as abuse.

        Returns:
            dict[str, Any]: The updated entity with the abuse flag.

        Raises:
            ValueError: If user ID or entity is not provided.
        """
        user = Users().get(user_id)
        entity = Contents().get(entity_id)
        if not (user and entity):
            raise ValueError("User ID or entity is not provided")
        abuse_flaggers = entity["abuse_flaggers"]
        first_flag_added = False
        if user["_id"] not in abuse_flaggers:
            abuse_flaggers.append(user["_id"])
            first_flag_added = len(abuse_flaggers) == 1
            Contents().update(
                entity["_id"],
                abuse_flaggers=abuse_flaggers,
            )
        if first_flag_added:
            cls.update_stats_for_course(
                entity["author_id"],
                entity["course_id"],
                active_flags=1,
            )
        updated_content = Contents().get(entity["_id"])
        if not updated_content:
            raise ValueError("Entity not found")
        return updated_content

    @classmethod
    def update_stats_after_unflag(
        cls, user_id: str, entity_id: str, has_no_historical_flags: bool, **kwargs: Any
    ) -> None:
        """Update the stats for the course after unflagging an entity."""
        entity = Contents().get(entity_id)
        if not entity:
            raise ObjectDoesNotExist

        first_historical_flag = (
            has_no_historical_flags and not entity["historical_abuse_flaggers"]
        )
        if first_historical_flag:
            cls.update_stats_for_course(user_id, entity["course_id"], inactive_flags=1)

        if not entity["abuse_flaggers"]:
            cls.update_stats_for_course(user_id, entity["course_id"], active_flags=-1)

    @classmethod
    def un_flag_as_abuse(
        cls, user_id: str, entity_id: str, **kwargs: Any
    ) -> dict[str, Any]:
        """
        Unflag an entity as abuse.

        Args:
            user (dict[str, Any]): The user who is unflagging the entity as abuse.
            entity (dict[str, Any]): The entity being unflagged as abuse.

        Returns:
            dict[str, Any]: The updated entity with the abuse flag removed.

        Raises:
            ValueError: If user ID or entity is not provided.
        """
        user = Users().get(user_id)
        entity = Contents().get(entity_id)
        if not (user and entity):
            raise ValueError("User ID or entity is not provided")

        has_no_historical_flags = len(entity["historical_abuse_flaggers"]) == 0
        if user["_id"] in entity["abuse_flaggers"]:
            entity["abuse_flaggers"].remove(user["_id"])
            Contents().update(
                entity["_id"],
                abuse_flaggers=entity["abuse_flaggers"],
            )
            cls.update_stats_after_unflag(
                entity["author_id"], entity["_id"], has_no_historical_flags
            )
        updated_content = Contents().get(entity["_id"])
        if not updated_content:
            raise ValueError("Entity not found")
        return updated_content

    @classmethod
    def un_flag_all_as_abuse(cls, entity_id: str, **kwargs: Any) -> dict[str, Any]:
        """
        Unflag an entity as abuse for all users.

        Args:
            entity (dict[str, Any]): The entity being unflagged as abuse.

        Returns:
            dict[str, Any]: The updated entity with all abuse flags removed.

        Raises:
            ValueError: If entity is not provided.
        """
        entity = Contents().get(entity_id)
        if not entity:
            raise ValueError("Entity is not provided")
        has_no_historical_flags = len(entity["historical_abuse_flaggers"]) == 0
        historical_abuse_flaggers = list(
            set(entity["historical_abuse_flaggers"]) | set(entity["abuse_flaggers"])
        )
        Contents().update(
            entity["_id"],
            abuse_flaggers=[],
            historical_abuse_flaggers=historical_abuse_flaggers,
        )
        cls.update_stats_after_unflag(
            entity["author_id"], entity["_id"], has_no_historical_flags
        )
        updated_content = Contents().get(entity["_id"])
        if not updated_content:
            raise ValueError("Entity not found")
        return updated_content

    @staticmethod
    def update_vote(
        content_id: str,
        user_id: str,
        vote_type: str = "",
        is_deleted: bool = False,
        **kwargs: Any,
    ) -> bool:
        """
        Update a vote on a thread (either upvote or downvote).

        :param content: The content document containing vote data.
        :param user: The user document for the user voting.
        :param vote_type: String indicating the type of vote ('up' or 'down').
        :param is_deleted: Boolean indicating if the user is removing their vote (True) or voting (False).
        :return: True if the vote was successfully updated, False otherwise.
        """
        user = Users().get(user_id)
        content = Contents().get(content_id)
        if not (user and content):
            raise ValueError("User ID or entity is not provided")

        votes: dict[str, Any] = content["votes"]
        update_needed: bool = False

        if not is_deleted:
            if vote_type not in ["up", "down"]:
                raise ValueError("Invalid vote_type, use ('up' or 'down')")

            if vote_type == "up":
                current_votes = set(votes["up"])
                opposite_votes = set(votes["down"])
            else:
                current_votes = set(votes["down"])
                opposite_votes = set(votes["up"])

            # Check if user is voting
            if user_id not in current_votes:
                current_votes.add(user_id)
                update_needed = True
                if user_id in opposite_votes:
                    opposite_votes.remove(user_id)

            updated_up_votes = opposite_votes if vote_type == "down" else current_votes
            updated_down_votes = (
                current_votes if vote_type == "down" else opposite_votes
            )

        else:
            # Handle vote deletion
            updated_up_votes = set(votes["up"])
            updated_down_votes = set(votes["down"])

            if user_id in updated_up_votes:
                updated_up_votes.remove(user_id)
                update_needed = True
            if user_id in updated_down_votes:
                updated_down_votes.remove(user_id)
                update_needed = True

        if update_needed:
            # Prepare updated votes
            content_model = Contents()
            updated_votes = content_model.get_votes_dict(
                list(updated_up_votes), list(updated_down_votes)
            )
            updated_count = content_model.update_votes(
                content_id=content_id, votes=updated_votes
            )
            return bool(updated_count)

        return False

    @classmethod
    def upvote_content(cls, entity_id: str, user_id: str, **kwargs: Any) -> bool:
        """
        Upvotes the specified thread or comment by the given user.

        Args:
            thread (dict): The thread or comment data to be upvoted.
            user (dict): The user who is performing the upvote.

        Returns:
            bool: True if the vote was successfully updated, False otherwise.
        """
        user = Users().get(user_id)
        entity = Contents().get(entity_id)
        if not (user and entity):
            raise ValueError("User ID or entity is not provided")

        return cls.update_vote(entity["_id"], user["external_id"], vote_type="up")

    @classmethod
    def downvote_content(cls, entity_id: str, user_id: str, **kwargs: Any) -> bool:
        """
        Downvotes the specified thread or comment by the given user.

        Args:
            thread (dict): The thread or comment data to be downvoted.
            user (dict): The user who is performing the downvote.

        Returns:
            bool: True if the vote was successfully updated, False otherwise.
        """
        user = Users().get(user_id)
        entity = Contents().get(entity_id)
        if not (user and entity):
            raise ValueError("User ID or entity is not provided")

        return cls.update_vote(entity["_id"], user["external_id"], vote_type="down")

    @classmethod
    def remove_vote(cls, entity_id: str, user_id: str, **kwargs: Any) -> bool:
        """
        Remove the vote (upvote or downvote) from the specified thread or comment for the given user.

        Args:
            thread (dict): The thread or comment data from which the vote should be removed.
            user (dict): The user who is removing their vote.

        Returns:
            bool: True if the vote was successfully removed, False otherwise.
        """
        user = Users().get(user_id)
        entity = Contents().get(entity_id)
        if not (user and entity):
            raise ValueError("User ID or entity is not provided")

        return cls.update_vote(entity["_id"], user["external_id"], is_deleted=True)

    @staticmethod
    def validate_thread_and_user(
        user_id: str, thread_id: str
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """
        Validate thread and user.

        Arguments:
            user_id (str): The ID of the user making the request.
            thread_id (str): The ID of the thread.

        Returns:
            tuple[dict[str, Any], dict[str, Any]]: A tuple containing the user and thread data.

        Raises:
            ValueError: If the thread or user is not found.
        """
        thread = CommentThread().get(thread_id)
        user = Users().get(user_id)
        if not (thread and user):
            raise ValueError("User / Thread doesn't exist")

        return user, thread

    @staticmethod
    def pin_unpin_thread(thread_id: str, action: str) -> None:
        """
        Pin or unpin the thread based on action parameter.

        Arguments:
            thread_id (str): The ID of the thread to pin/unpin.
            action (str): The action to perform ("pin" or "unpin").
        """
        CommentThread().update(
            thread_id, pinned=action == "pin", skip_timestamp_update=True
        )

    @classmethod
    def get_pinned_unpinned_thread_serialized_data(
        cls, user_id: str, thread_id: str, serializer_class: Any
    ) -> dict[str, Any]:
        """
        Return serialized data of pinned or unpinned thread.

        Arguments:
            user (dict[str, Any]): The user who requested the action.
            thread_id (str): The ID of the thread to pin/unpin.

        Returns:
            dict[str, Any]: The serialized data of the pinned/unpinned thread.

        Raises:
            ValueError: If the serialization is not valid.
        """
        user = Users().get(user_id)
        updated_thread = CommentThread().get(thread_id)
        if not (user and updated_thread):
            raise ValueError("User ID or entity is not provided")

        context = {
            "user_id": user["_id"],
            "username": user["username"],
            "type": "thread",
            "id": thread_id,
        }
        if updated_thread is not None:
            context = {**context, **updated_thread}
        serializer = serializer_class(data=context, backend=cls)
        if not serializer.is_valid():
            raise ValueError(serializer.errors)

        return serializer.data

    @classmethod
    def handle_pin_unpin_thread_request(
        cls, user_id: str, thread_id: str, action: str, serializer_class: Any
    ) -> dict[str, Any]:
        """
        Catches pin/unpin thread request.

        - validates thread and user.
        - pin or unpin the thread based on action parameter.
        - return serialized data of thread.

        Arguments:
            user_id (str): The ID of the user making the request.
            thread_id (str): The ID of the thread to pin/unpin.
            action (str): The action to perform ("pin" or "unpin").

        Returns:
            dict[str, Any]: The serialized data of the pinned/unpinned thread.
        """
        user, _ = cls.validate_thread_and_user(user_id, thread_id)
        cls.pin_unpin_thread(thread_id, action)
        return cls.get_pinned_unpinned_thread_serialized_data(
            user["external_id"], thread_id, serializer_class
        )

    @staticmethod
    def get_abuse_flagged_count(thread_ids: list[str]) -> dict[str, int]:
        """
        Retrieves the count of abuse-flagged comments for each thread in the provided list of thread IDs.

        Args:
            thread_ids (list[str]): List of thread IDs to check for abuse flags.

        Returns:
            dict[str, int]: A dictionary mapping thread IDs to their corresponding abuse-flagged comment count.
        """
        pipeline: list[dict[str, Any]] = [
            {
                "$match": {
                    "comment_thread_id": {"$in": [ObjectId(tid) for tid in thread_ids]},
                    "abuse_flaggers": {"$ne": []},
                }
            },
            {"$group": {"_id": "$comment_thread_id", "flagged_count": {"$sum": 1}}},
        ]
        flagged_threads = Contents().aggregate(pipeline)

        return {str(item["_id"]): item["flagged_count"] for item in flagged_threads}

    @staticmethod
    def get_read_states(
        thread_ids: list[str], user_id: str, course_id: str
    ) -> dict[str, list[Any]]:
        """
        Retrieves the read state and unread comment count for each thread in the provided list.

        Args:
            threads (list[dict[str, Any]]): list of threads to check read state for.
            user_id (str): The ID of the user whose read states are being retrieved.
            course_id (str): The course ID associated with the threads.

        Returns:
            dict[str, list[Any]]: A dictionary mapping thread IDs to a list containing
            whether the thread is read and the unread comment count.
        """
        threads = CommentThread().find(
            {"_id": {"$in": [ObjectId(thread_id) for thread_id in thread_ids]}}
        )
        read_states = {}
        if user_id:
            user = Users().find_one(
                {"_id": user_id, "read_states.course_id": course_id}
            )
            if not user:
                return {}

            read_state = next(
                (
                    rs
                    for rs in user.get("read_states", [])
                    if rs.get("course_id") == course_id
                ),
                None,
            )
            if read_state:
                read_dates = read_state.get("last_read_times", {})
                for thread in threads:
                    thread_key = str(thread["_id"])
                    if thread_key in read_dates:
                        read_date = make_aware(read_dates[thread_key])
                        last_activity_at = make_aware(thread["last_activity_at"])
                        is_read = read_date >= last_activity_at
                        unread_comment_count = Contents().count_documents(
                            {
                                "comment_thread_id": ObjectId(thread_key),
                                "created_at": {"$gte": read_dates[thread_key]},
                                "author_id": {"$ne": str(user_id)},
                            }
                        )
                        read_states[thread_key] = [is_read, unread_comment_count]

        return read_states

    @staticmethod
    def get_endorsed(thread_ids: list[str]) -> dict[str, bool]:
        """
        Retrieves endorsed status for each thread in the provided list of thread IDs.

        Args:
            thread_ids (list[str]): List of thread IDs to check for endorsement.

        Returns:
            dict[str, bool]: A dictionary of thread IDs to their endorsed status (True if endorsed, False otherwise).
        """
        endorsed_comments = Comment().find(
            {
                "comment_thread_id": {"$in": [ObjectId(tid) for tid in thread_ids]},
                "endorsed": True,
            }
        )

        return {str(item["comment_thread_id"]): True for item in endorsed_comments}

    @staticmethod
    def get_user_read_state_by_course_id(
        user_id: str, course_id: str
    ) -> dict[str, Any]:
        """
        Retrieves the user's read state for a specific course.

        Args:
            user (dict[str, Any]): The user object containing read states.
            course_id (str): The course ID to filter the user's read state by.

        Returns:
            dict[str, Any]: The user's read state for the specified course, or an empty dictionary if not found.
        """
        user = Users().get(user_id)
        if not user:
            raise ValueError("User does not exist.")

        for read_state in user.get("read_states", []):
            if read_state["course_id"] == course_id:
                return read_state
        return {}

    # TODO: Make this function modular
    # pylint: disable=too-many-nested-blocks,too-many-statements
    @classmethod
    def handle_threads_query(
        cls,
        comment_thread_ids: list[str],
        user_id: str,
        course_id: str,
        group_ids: list[int],
        author_id: Optional[str],
        thread_type: Optional[str],
        filter_flagged: bool,
        filter_unread: bool,
        filter_unanswered: bool,
        filter_unresponded: bool,
        count_flagged: bool,
        sort_key: str,
        page: int,
        per_page: int,
        context: str = "course",
        raw_query: bool = False,
        commentable_ids: Optional[list[str]] = None,
        is_moderator: bool = False,
    ) -> dict[str, Any]:
        """
        Handles complex thread queries based on various filters and returns paginated results.

        Args:
            comment_thread_ids (list[str]): List of comment thread IDs to filter.
            user_id (str): The ID of the user making the request.
            course_id (str): The course ID associated with the threads.
            group_ids (list[int]): List of group IDs for group-based filtering.
            author_id (str): The ID of the author to filter threads by.
            thread_type (str): The type of thread to filter by.
            filter_flagged (bool): Whether to filter threads flagged for abuse.
            filter_unread (bool): Whether to filter unread threads.
            filter_unanswered (bool): Whether to filter unanswered questions.
            filter_unresponded (bool): Whether to filter threads with no responses.
            count_flagged (bool): Whether to include flagged content count.
            sort_key (str): The key to sort the threads by.
            page (int): The page number for pagination.
            per_page (int): The number of threads per page.
            context (str): The context to filter threads by.
            raw_query (bool): Whether to return raw query results without further processing.
            commentable_ids (Optional[list[str]]): List of commentable IDs to filter threads by topic id.
            is_moderator (bool): Whether the user is a discussion moderator.

        Returns:
            dict[str, Any]: A dictionary containing the paginated thread results and associated metadata.
        """
        # Convert thread_ids to ObjectId
        comment_thread_obj_ids: list[ObjectId] = []

        for tid in comment_thread_ids:
            try:
                thread_id = ObjectId(tid)
                comment_thread_obj_ids.append(thread_id)
            except bson_errors.InvalidId:
                continue

        # Base query
        base_query: dict[str, Any] = {
            "_id": {"$in": comment_thread_obj_ids},
            "context": context,
        }

        # Group filtering
        if group_ids:
            base_query["$or"] = [
                {"group_id": {"$in": group_ids}},
                {"group_id": {"$exists": False}},
            ]

        # Author filtering
        if author_id:
            base_query["author_id"] = author_id
            if author_id != user_id:
                base_query["anonymous"] = False
                base_query["anonymous_to_peers"] = False

        # Thread type filtering
        if thread_type:
            base_query["thread_type"] = thread_type

        # Flagged content filtering
        if filter_flagged:
            flagged_query = {
                "course_id": course_id,
                "abuse_flaggers": {"$ne": [], "$exists": True},
            }
            flagged_comments = Comment().distinct("comment_thread_id", flagged_query)
            flagged_threads = CommentThread().distinct("_id", flagged_query)
            base_query["_id"]["$in"] = list(
                set(comment_thread_obj_ids) & set(flagged_comments + flagged_threads)
            )

        # Unanswered questions filtering
        if filter_unanswered:
            endorsed_threads = Comment().distinct(
                "comment_thread_id",
                {
                    "course_id": course_id,
                    "parent_id": {"$exists": False},
                    "endorsed": True,
                },
            )
            base_query["thread_type"] = "question"
            base_query["_id"]["$nin"] = endorsed_threads

        # Unresponded threads filtering
        if filter_unresponded:
            base_query["comment_count"] = 0

        # filter by topics: if commentable_ids are provided, commentable_id is basically topic id
        # For moderators: show all topics (no filtering by commentable_ids)
        # For learners: apply commentable_ids filtering (cohorted topics shown as archived)
        if commentable_ids and not is_moderator:
            base_query["commentable_id"] = {"$in": commentable_ids}
        sort_criteria = get_sort_criteria(sort_key)

        comment_threads = CommentThread().find(base_query)
        thread_count = CommentThread().count_documents(base_query)

        if sort_criteria or raw_query:
            request_user = Users().get(user_id) if user_id else None

            if not raw_query:
                comment_threads = comment_threads.sort(sort_criteria)

            if filter_unread and request_user:
                read_state = cls.get_user_read_state_by_course_id(
                    request_user["external_id"], course_id
                )
                read_dates = read_state.get("last_read_times", {})

                threads = []
                skipped = 0
                to_skip = (page - 1) * per_page
                has_more = False
                batch_size = 100

                for thread in comment_threads.batch_size(batch_size):
                    thread_key = str(thread["_id"])
                    if (
                        thread_key not in read_dates
                        or read_dates[thread_key] < thread["last_activity_at"]
                    ):
                        if raw_query:
                            threads.append(thread)
                        elif skipped >= to_skip:
                            if len(threads) == per_page:
                                has_more = True
                                break
                            threads.append(thread)
                        else:
                            skipped += 1
                num_pages = page + 1 if has_more else page
            else:
                if raw_query:
                    threads = list(comment_threads)
                else:
                    page = max(1, page)
                    paginated_collection = comment_threads.skip(
                        (page - 1) * per_page
                    ).limit(per_page)
                    threads = list(paginated_collection)
                    num_pages = max(1, math.ceil(thread_count / per_page))

            if raw_query:
                return {"result": threads}
            if len(threads) == 0:
                collection = []
            else:
                thread_ids = [str(thread["_id"]) for thread in threads]
                collection = cls.threads_presentor(
                    thread_ids, user_id, course_id, count_flagged
                )

            return {
                "collection": collection,
                "num_pages": num_pages,
                "page": page,
                "thread_count": thread_count,
            }

        return {}

    @staticmethod
    def prepare_thread(
        thread_id: str,
        is_read: bool,
        unread_count: int,
        is_endorsed: bool,
        abuse_flagged_count: int,
    ) -> dict[str, Any]:
        """
        Prepares thread data for presentation.

        Args:
            thread (dict[str, Any]): The thread data.
            is_read (bool): Whether the thread is read.
            unread_count (int): The count of unread comments.
            is_endorsed (bool): Whether the thread is endorsed.
            abuse_flagged_count (int): The abuse flagged count.

        Returns:
            dict[str, Any]: A dictionary representing the prepared thread data.
        """
        thread = CommentThread().get(thread_id)
        if not thread:
            raise ValueError("Thread does not exist.")

        return {
            "id": str(thread["_id"]),
            **thread,
            "type": "thread",
            "read": is_read,
            "unread_comments_count": unread_count,
            "endorsed": is_endorsed,
            "abuse_flagged_count": abuse_flagged_count,
        }

    @classmethod
    def threads_presentor(
        cls,
        thread_ids: list[str],
        user_id: str,
        course_id: str,
        count_flagged: bool = False,
    ) -> list[dict[str, Any]]:
        """
        Presents the threads by preparing them for display.

        Args:
            threads (list[dict[str, Any]]): List of threads to present.
            user_id (str): The ID of the user presenting the threads.
            course_id (str): The course ID associated with the threads.
            count_flagged (bool, optional): Whether to include flagged content count. Defaults to False.

        Returns:
            list[dict[str, Any]]: A list of prepared thread data.
        """
        threads = CommentThread().find(
            {"_id": {"$in": [ObjectId(thread_id) for thread_id in thread_ids]}}
        )
        read_states = cls.get_read_states(thread_ids, user_id, course_id)
        threads_endorsed = cls.get_endorsed(thread_ids)
        threads_flagged = (
            cls.get_abuse_flagged_count(thread_ids) if count_flagged else {}
        )
        threads_dict = {str(thread["_id"]): thread for thread in threads}

        presenters = []
        for thread_id in thread_ids:
            thread = threads_dict.get(thread_id)
            if thread:
                thread_key = thread_id
                is_read, unread_count = read_states.get(
                    thread_key, (False, thread["comment_count"])
                )
                is_endorsed = threads_endorsed.get(thread_key, False)
                abuse_flagged_count = threads_flagged.get(thread_key, 0)
                presenters.append(
                    cls.prepare_thread(
                        thread["_id"],
                        is_read,
                        unread_count,
                        is_endorsed,
                        abuse_flagged_count,
                    )
                )

        return presenters

    @staticmethod
    def get_username_from_id(user_id: str) -> Optional[str]:
        """
        Retrieve the username associated with a given user ID.

        Args:
            _id (int): The unique identifier of the user.

        Returns:
            Optional[str]: The username of the user if found, or None if not.

        """
        user = Users().get(_id=user_id) or {}
        if username := user.get("username"):
            return username
        return None

    @staticmethod
    def validate_object(model: str, obj_id: str) -> Any:
        """
        Validates the object if it exists or not.

        Parameters:
            model: The model for which to validate the id.
            id: The ID of the object to validate in the model.
        Response:
            raise exception if object does not exists.
            return object
        """
        models = {
            "Comment": Comment,
            "CommentThread": CommentThread,
        }
        instance = models[model]().get(obj_id)
        if not instance:
            raise ObjectDoesNotExist
        return instance

    @staticmethod
    def find_subscribed_threads(
        user_id: str, course_id: Optional[str] = None
    ) -> list[str]:
        """
        Find threads that a user is subscribed to in a specific course.

        Args:
            user_id (str): The ID of the user.
            course_id (str): The ID of the course.

        Returns:
            list: A list of thread ids that the user is subscribed to in the course.
        """
        subscriptions = Subscriptions()
        threads = CommentThread()

        subscription_filter = {"subscriber_id": user_id}
        subscriptions_cursor = subscriptions.find(subscription_filter)

        thread_ids = []
        for subscription in subscriptions_cursor:
            thread_ids.append(ObjectId(subscription["source_id"]))

        thread_filter: dict[str, Any] = {"_id": {"$in": thread_ids}}
        if course_id:
            thread_filter["course_id"] = course_id
        threads_cursor = threads.find(thread_filter)

        subscribed_ids = []
        for thread in threads_cursor:
            subscribed_ids.append(str(thread["_id"]))

        return subscribed_ids

    @staticmethod
    def subscribe_user(
        user_id: str, source_id: str, source_type: str
    ) -> dict[str, Any] | None:
        """Subscribe a user to a source."""
        subscription = Subscriptions().get_subscription(user_id, source_id)
        if not subscription:
            Subscriptions().insert(user_id, source_id, source_type)
            subscription = Subscriptions().get_subscription(user_id, source_id)
        if subscription:
            subscription["_id"] = str(subscription["_id"])
        return subscription

    @staticmethod
    def unsubscribe_user(
        user_id: str, source_id: str, source_type: Optional[str] = ""
    ) -> None:
        """Unsubscribe a user from a source."""
        Subscriptions().delete_subscription(user_id, source_id, source_type=source_type)

    @staticmethod
    def delete_comments_of_a_thread(thread_id: str) -> None:
        """Delete comments of a thread."""
        for comment in Comment().get_list(
            comment_thread_id=ObjectId(thread_id),
            depth=0,
            parent_id=None,
        ):
            Comment().delete(comment["_id"])

    @staticmethod
    def delete_subscriptions_of_a_thread(thread_id: str) -> None:
        """Delete subscriptions of a thread."""
        for subscription in Subscriptions().get_list(
            source_id=thread_id, source_type="CommentThread"
        ):
            Subscriptions().delete_subscription(
                subscription["subscriber_id"],
                subscription["source_id"],
                source_type="CommentThread",
            )

    @staticmethod
    def validate_params(params: dict[str, Any], user_id: Optional[str] = None) -> None:
        """
        Validate the request parameters.
        Args:
            params (dict): The request parameters.
            user_id (optional[str]): The Id of the user for validation.

        Returns:
            Response: A Response object with an error message if doesn't exist.
        """
        valid_params = [
            "course_id",
            "author_id",
            "thread_type",
            "flagged",
            "unread",
            "unanswered",
            "unresponded",
            "count_flagged",
            "sort_key",
            "page",
            "per_page",
            "request_id",
            "commentable_ids",
            "group_id",
            "group_ids",
        ]
        if not user_id:
            valid_params.append("user_id")
            user_id = params.get("user_id")

        for key in params:
            if key not in valid_params:
                raise ForumV2RequestError(f"Invalid parameter: {key}")

        if "course_id" not in params:
            raise ForumV2RequestError("Missing required parameter: course_id")

        if user_id:
            user = Users().get(user_id)
            if not user:
                raise ForumV2RequestError("User doesn't exist")

    @classmethod
    def get_threads(
        cls,
        params: dict[str, Any],
        user_id: str,
        serializer: Any,
        thread_ids: list[str],
    ) -> dict[str, Any]:
        """get subscribed or all threads of a specific course for a specific user."""
        count_flagged = bool(params.get("count_flagged", False))
        threads = cls.handle_threads_query(
            thread_ids,
            user_id,
            params["course_id"],
            get_group_ids_from_params(params),
            params.get("author_id", ""),
            params.get("thread_type"),
            bool(params.get("flagged", False)),
            bool(params.get("unread", False)),
            bool(params.get("unanswered", False)),
            bool(params.get("unresponded", False)),
            count_flagged,
            params.get("sort_key", ""),
            int(params.get("page", 1)),
            int(params.get("per_page", 100)),
            commentable_ids=params.get("commentable_ids", []),
            is_moderator=params.get("is_moderator", False),
        )
        context: dict[str, Any] = {
            "count_flagged": count_flagged,
            "include_endorsed": True,
            "include_read_state": True,
        }
        if user_id:
            context["user_id"] = user_id
        serializer = serializer(
            threads.pop("collection"), many=True, context=context, backend=cls
        )
        threads["collection"] = serializer.data
        return threads

    @staticmethod
    def generate_id() -> str:
        return str(ObjectId())

    @staticmethod
    def find_or_create_user(
        user_id: str, username: Optional[str] = "", default_sort_key: Optional[str] = ""
    ) -> str:
        """Find or create user."""
        user = Users().get(user_id)
        if user:
            return user["external_id"]
        user_id = Users().insert(
            user_id, username=username, default_sort_key=default_sort_key
        )
        return user_id

    @classmethod
    def create_comment(cls, data: dict[str, Any]) -> str:
        """
        handle comment creation and returns a comment.

        Parameters:
            data: The content of the comment.

        Response:
            The details of the comment that is created.
        """
        new_comment_id = Comment().insert(
            body=data["body"],
            author_id=data["author_id"],
            author_username=data.get("author_username"),
            course_id=data["course_id"],
            anonymous=data.get("anonymous", False),
            anonymous_to_peers=data.get("anonymous_to_peers", False),
            depth=data.get("depth", 0),
            comment_thread_id=data["comment_thread_id"],
            parent_id=data.get("parent_id"),
        )

        if data.get("parent_id"):
            cls.update_stats_for_course(data["author_id"], data["course_id"], replies=1)
        else:
            cls.update_stats_for_course(
                data["author_id"], data["course_id"], responses=1
            )

        return str(new_comment_id)

    @staticmethod
    def update_comment_and_get_updated_comment(
        comment_id: str,
        body: Optional[str] = None,
        course_id: Optional[str] = None,
        user_id: Optional[str] = None,
        anonymous: Optional[bool] = False,
        anonymous_to_peers: Optional[bool] = False,
        endorsed: Optional[bool] = None,
        closed: Optional[bool] = False,
        editing_user_id: Optional[str] = None,
        edit_reason_code: Optional[str] = None,
        endorsement_user_id: Optional[str] = None,
    ) -> dict[str, Any] | None:
        """
        Update an existing child/parent comment.

        Parameters:
            comment_id: The ID of the comment to be edited.
            body (Optional[str]): The content of the comment.
            course_id (Optional[str]): The Id of the respective course.
            user_id (Optional[str]): The requesting user id.
            anonymous (Optional[bool]): anonymous flag(True or False).
            anonymous_to_peers (Optional[bool]): anonymous to peers flag(True or False).
            endorsed (Optional[bool]): Flag indicating if the comment is endorsed by any user.
            closed (Optional[bool]): Flag indicating if the comment thread is closed.
            editing_user_id (Optional[str]): The ID of the user editing the comment.
            edit_reason_code (Optional[str]): The reason for editing the comment, typically represented by a code.
            endorsement_user_id (Optional[str]): The ID of the user endorsing the comment.
        Response:
            The details of the comment that is updated.
        """
        Comment().update(
            comment_id,
            body=body,
            course_id=course_id,
            author_id=user_id,
            anonymous=anonymous,
            anonymous_to_peers=anonymous_to_peers,
            endorsed=endorsed,
            closed=closed,
            editing_user_id=editing_user_id,
            edit_reason_code=edit_reason_code,
            endorsement_user_id=endorsement_user_id,
        )
        return Comment().get(comment_id)

    @staticmethod
    def get_commentables_counts_based_on_type(course_id: str) -> dict[str, Any]:
        """Return commentables counts in a course based on thread's type."""
        pipeline: list[dict[str, Any]] = [
            {"$match": {"course_id": course_id, "_type": "CommentThread"}},
            {
                "$group": {
                    "_id": {"topic_id": "$commentable_id", "type": "$thread_type"},
                    "count": {"$sum": 1},
                }
            },
        ]

        result = CommentThread().aggregate(pipeline)
        commentable_counts = {}
        for commentable in result:
            topic_id = commentable["_id"]["topic_id"]
            if topic_id not in commentable_counts:
                commentable_counts[topic_id] = {"discussion": 0, "question": 0}
            commentable_counts[topic_id].update(
                {commentable["_id"]["type"]: commentable["count"]}
            )

        return commentable_counts

    @classmethod
    def get_user_voted_ids(cls, user_id: str, vote: str) -> list[str]:
        """Get the IDs of the posts voted by a user."""
        if vote not in ["up", "down"]:
            raise ValueError("Invalid vote type")

        content_model = Contents()
        contents = content_model.get_list()
        voted_ids = []
        for content in contents:
            votes = content["votes"][vote]
            if user_id in votes:
                voted_ids.append(content["_id"])

        return voted_ids

    @staticmethod
    def filter_standalone_threads(comment_ids: list[str]) -> list[str]:
        """Filter out standalone threads from the list of threads."""
        comments = Comment().find({"_id": {"$in": comment_ids}})
        filtered_comments = []
        for comment in comments:
            if not comment["context"] == "standalone":
                filtered_comments.append(comment)
        return [str([comment["comment_thread_id"]]) for comment in filtered_comments]

    @classmethod
    def user_to_hash(
        cls, user_id: str, params: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """
        Converts user data to a hash
        """
        user = Users().get(user_id)
        if not user:
            raise ValueError("User not found.")
        if params is None:
            params = {}

        hash_data = {}
        hash_data["username"] = user["username"]
        hash_data["external_id"] = user["external_id"]
        hash_data["id"] = user["external_id"]

        comment_model = Comment()
        thread_model = CommentThread()

        if params.get("complete"):
            subscribed_thread_ids = cls.find_subscribed_threads(user["external_id"])
            upvoted_ids = cls.get_user_voted_ids(user["external_id"], "up")
            downvoted_ids = cls.get_user_voted_ids(user["external_id"], "down")
            hash_data.update(
                {
                    "subscribed_thread_ids": subscribed_thread_ids,
                    "subscribed_commentable_ids": [],
                    "subscribed_user_ids": [],
                    "follower_ids": [],
                    "id": user["external_id"],
                    "upvoted_ids": upvoted_ids,
                    "downvoted_ids": downvoted_ids,
                    "default_sort_key": user["default_sort_key"],
                }
            )

        if params.get("course_id"):
            threads = thread_model.find(
                {
                    "author_id": user["external_id"],
                    "course_id": params["course_id"],
                    "anonymous": False,
                    "anonymouse_to_peers": False,
                }
            )
            comments = comment_model.find(
                {
                    "author_id": user["external_id"],
                    "course_id": params["course_id"],
                    "anonymous": False,
                    "anonymouse_to_peers": False,
                }
            )
            if params.get("group_ids"):
                specified_groups_or_global = params["group_ids"] + [None]
                group_query = {
                    "_id": {"$in": [thread["_id"] for thread in threads]},
                    "$and": [
                        {"group_id": {"$in": specified_groups_or_global}},
                        {"group_id": {"$exists": False}},
                    ],
                }
                group_threads = CommentThread().find(group_query)
                group_thread_ids = [str(thread["_id"]) for thread in group_threads]
                threads_count = len(group_thread_ids)
                comment_ids = [comment["_id"] for comment in comments]
                comment_thread_ids = cls.filter_standalone_threads(comment_ids)

                group_query = {
                    "_id": {"$in": [ObjectId(tid) for tid in comment_thread_ids]},
                    "$and": [
                        {"group_id": {"$in": specified_groups_or_global}},
                        {"group_id": {"$exists": False}},
                    ],
                }
                group_comment_threads = thread_model.find(group_query)
                group_comment_thread_ids = [
                    str(thread["_id"]) for thread in group_comment_threads
                ]
                comments_count = sum(
                    1
                    for comment_thread_id in comment_thread_ids
                    if comment_thread_id in group_comment_thread_ids
                )
            else:
                thread_ids = [str(thread["_id"]) for thread in threads]
                threads_count = len(thread_ids)
                comment_ids = [comment["_id"] for comment in comments]
                comment_thread_ids = cls.filter_standalone_threads(comment_ids)
                comments_count = len(comment_thread_ids)

            hash_data.update(
                {
                    "threads_count": threads_count,
                    "comments_count": comments_count,
                }
            )

        return hash_data

    @staticmethod
    def replace_username_in_all_content(user_id: str, username: str) -> None:
        """Replace new username in all content documents."""
        content_model = Contents()
        contents = content_model.get_list(author_id=user_id)
        for content in contents:
            content_model.update(
                content["_id"],
                author_username=username,
            )

    @staticmethod
    def unsubscribe_all(user_id: str) -> None:
        """Unsubscribe user from all content."""
        subscriptions = Subscriptions()
        subscription_filter = {"subscriber_id": user_id}
        subscriptions_cursor = subscriptions.find(subscription_filter)

        for subscription in subscriptions_cursor:
            subscriptions.delete(subscription["_id"])

    @staticmethod
    def retire_all_content(user_id: str, username: str) -> None:
        """Retire all content from user."""
        content_model = Contents()
        contents = content_model.get_list(author_id=user_id)
        for content in contents:
            content_model.update(
                content["_id"],
                author_username=username,
                body=RETIRED_BODY,
            )
            if content["_type"] == "CommentThread":
                content_model.update(
                    content["_id"],
                    title=RETIRED_TITLE,
                )

    @staticmethod
    def find_or_create_read_state(user_id: str, thread_id: str) -> dict[str, Any]:
        """Find or create user read states."""
        user = Users().get(user_id)
        if not user:
            raise ObjectDoesNotExist
        thread = CommentThread().get(thread_id)
        if not thread:
            raise ObjectDoesNotExist

        read_states = user.get("read_states", [])
        for state in read_states:
            if state["course_id"] == thread["course_id"]:
                return state

        read_state = {
            "_id": ObjectId(),
            "course_id": thread["course_id"],
            "last_read_times": {},
        }
        read_states.append(read_state)
        Users().update(user_id, read_states=read_states)
        return read_state

    @classmethod
    def mark_as_read(cls, user_id: str, thread_id: str) -> None:
        """Mark thread as read."""
        user = Users().get(user_id)
        thread = CommentThread().get(thread_id)
        if not (user and thread):
            raise ValueError("User and/or Thread not found.")
        read_state = cls.find_or_create_read_state(user["external_id"], thread["_id"])

        read_state["last_read_times"].update(
            {
                str(thread["_id"]): datetime.now(timezone.utc),
            }
        )
        update_user = Users().get(user["external_id"])
        if not update_user:
            raise ObjectDoesNotExist
        new_read_states = update_user["read_states"]
        updated_read_states = []
        for state in new_read_states:
            if state["course_id"] == thread["course_id"]:
                state = read_state
            updated_read_states.append(state)

        Users().update(user["external_id"], read_states=updated_read_states)

    @staticmethod
    def find_or_create_user_stats(user_id: str, course_id: str) -> dict[str, Any]:
        """Find or create user stats document."""
        user = Users().get(user_id)
        if not user:
            raise ObjectDoesNotExist

        course_stats = user.get("course_stats", [])
        for stat in course_stats:
            if stat["course_id"] == course_id:
                return stat

        course_stat = {
            "_id": ObjectId(),
            "active_flags": 0,
            "inactive_flags": 0,
            "threads": 0,
            "responses": 0,
            "replies": 0,
            "course_id": course_id,
            "last_activity_at": "",
        }
        course_stats.append(course_stat)
        Users().update(user["external_id"], course_stats=course_stats)
        return course_stat

    @staticmethod
    def update_user_stats_for_course(user_id: str, stat: dict[str, Any]) -> None:
        """Update user stats for course."""
        user = Users().get(user_id)
        if not user:
            raise ObjectDoesNotExist
        updated_course_stats = []
        course_stats = user["course_stats"]
        for course_stat in course_stats:
            if course_stat["course_id"] == stat["course_id"]:
                course_stat.update(stat)
            updated_course_stats.append(course_stat)
        Users().update(user_id, course_stats=updated_course_stats)

    @classmethod
    def build_course_stats(cls, author_id: str, course_id: str) -> None:
        """Build course stats."""
        user = Users().get(author_id)
        if not user:
            raise ObjectDoesNotExist
        pipeline = [
            {
                "$match": {
                    "course_id": course_id,
                    "author_id": user["external_id"],
                    "anonymous_to_peers": False,
                    "anonymous": False,
                }
            },
            {
                "$addFields": {
                    "is_reply": {"$ne": [{"$ifNull": ["$parent_id", None]}, None]}
                }
            },
            {
                "$group": {
                    "_id": {"type": "$_type", "is_reply": "$is_reply"},
                    "count": {"$sum": 1},
                    "active_flags": {
                        "$sum": {
                            "$cond": {
                                "if": {"$gt": [{"$size": "$abuse_flaggers"}, 0]},
                                "then": 1,
                                "else": 0,
                            }
                        }
                    },
                    "inactive_flags": {
                        "$sum": {
                            "$cond": {
                                "if": {
                                    "$gt": [{"$size": "$historical_abuse_flaggers"}, 0]
                                },
                                "then": 1,
                                "else": 0,
                            }
                        }
                    },
                    "latest_update_at": {"$max": "$updated_at"},
                }
            },
        ]

        data = list(Contents().aggregate(pipeline))
        active_flags = 0
        inactive_flags = 0
        threads = 0
        responses = 0
        replies = 0
        updated_at = datetime.utcfromtimestamp(0)

        for counts in data:
            _type, is_reply = counts["_id"]["type"], counts["_id"]["is_reply"]
            last_update_at = counts.get("latest_update_at", datetime(1970, 1, 1))
            if _type == "Comment" and is_reply:
                replies = counts["count"]
            elif _type == "Comment" and not is_reply:
                responses = counts["count"]
            else:
                threads = counts["count"]
            last_update_at = make_aware(last_update_at)
            updated_at = make_aware(updated_at)
            updated_at = max(last_update_at, updated_at)
            active_flags += counts["active_flags"]
            inactive_flags += counts["inactive_flags"]

        stats = cls.find_or_create_user_stats(user["external_id"], course_id)
        stats["replies"] = replies
        stats["responses"] = responses
        stats["threads"] = threads
        stats["active_flags"] = active_flags
        stats["inactive_flags"] = inactive_flags
        stats["last_activity_at"] = updated_at
        cls.update_user_stats_for_course(user["external_id"], stats)

    @classmethod
    def update_all_users_in_course(cls, course_id: str) -> list[str]:
        """Update all user stats in a course."""
        course_contents = Contents().get_list(
            anonymous=False,
            anonymous_to_peers=False,
            course_id=course_id,
        )
        author_ids = []
        for content in course_contents:
            if content["author_id"] not in author_ids:
                author_ids.append(content["author_id"])

        for author_id in author_ids:
            cls.build_course_stats(author_id, course_id)
        return author_ids

    @staticmethod
    def get_user_by_username(username: str | None) -> dict[str, Any] | None:
        """Return user from username."""
        cursor = Users().find({"username": username})
        try:
            return next(cursor)
        except StopIteration:
            return None

    @staticmethod
    def get_comment(comment_id: str) -> dict[str, Any] | None:
        """Get comment from id."""
        comment = Comment().get(comment_id)
        return comment

    @staticmethod
    def get_thread(thread_id: str) -> dict[str, Any] | None:
        """Get thread from id."""
        thread = CommentThread().get(thread_id)
        if not thread:
            return None
        return thread

    @staticmethod
    def get_comments(**kwargs: Any) -> list[dict[str, Any]]:
        """Return comments from kwargs."""
        if "comment_thread_id" in kwargs:
            kwargs["comment_thread_id"] = ObjectId(kwargs["comment_thread_id"])
        if parent_id := kwargs.get("parent_id"):
            kwargs["parent_id"] = ObjectId(parent_id)

        return list(Comment().get_list(**kwargs))

    @staticmethod
    def get_comments_count(**kwargs: Any) -> int:
        """Return comments from kwargs."""
        if "comment_thread_id" in kwargs:
            kwargs["comment_thread_id"] = ObjectId(kwargs["comment_thread_id"])
        if parent_id := kwargs.get("parent_id"):
            kwargs["parent_id"] = ObjectId(parent_id)

        return Comment().count_documents(kwargs)

    @staticmethod
    def update_comment(comment_id: str, **kwargs: Any) -> int:
        """Update comment."""
        return Comment().update(comment_id, **kwargs)

    @staticmethod
    def delete_comment(comment_id: str) -> None:
        """Delete comment."""
        Comment().delete(comment_id)

    @staticmethod
    def get_thread_id_from_comment(comment_id: str) -> dict[str, Any] | None:
        """Return thread_id from comment_id."""
        parent_comment = Comment().get(comment_id)
        if parent_comment:
            return parent_comment["comment_thread_id"]
        raise ValueError("Comment doesn't have the thread.")

    @staticmethod
    def get_user(user_id: str, get_full_dict: bool = True) -> dict[str, Any] | None:
        """Return user from user_id."""
        return Users().get(user_id)

    @staticmethod
    def get_subscription(
        subscriber_id: str, source_id: str, **kwargs: Any
    ) -> dict[str, Any] | None:
        """Return subscription from subscriber_id and source_id."""
        subscription = Subscriptions().get_subscription(subscriber_id, source_id)
        if not subscription:
            return None
        return subscription

    @staticmethod
    def get_subscriptions(query: dict[str, Any]) -> list[dict[str, Any]]:
        """Return subscriptions from filter."""
        return list(Subscriptions().find(query))

    @staticmethod
    def delete_thread(thread_id: str) -> int:
        """Delete thread."""
        return CommentThread().delete(thread_id)

    @staticmethod
    def create_thread(data: dict[str, Any]) -> str:
        """Create thread."""
        new_thread_id = CommentThread().insert(
            title=data["title"],
            body=data["body"],
            course_id=data["course_id"],
            anonymous=str_to_bool(data.get("anonymous", "False")),
            anonymous_to_peers=str_to_bool(data.get("anonymous_to_peers", "False")),
            author_id=data["author_id"],
            commentable_id=data.get("commentable_id", "course"),
            thread_type=data.get("thread_type", "discussion"),
            author_username=data.get("author_username"),
            context=data.get("context", "course"),
            pinned=data.get("pinned", False),
            visible=data.get("visible", True),
            abuse_flaggers=data.get("abuse_flaggers"),
            historical_abuse_flaggers=data.get("historical_abuse_flaggers"),
            group_id=data.get("group_id"),
        )
        return new_thread_id

    @staticmethod
    def update_thread(thread_id: str, **kwargs: Any) -> int:
        """Update thread."""
        return CommentThread().update(thread_id, **kwargs)

    @staticmethod
    def get_filtered_threads(
        query: dict[str, Any], ids_only: bool = False
    ) -> list[dict[str, Any]]:
        """Return threads from filter."""
        thread_filter = {
            "_type": {"$in": [CommentThread().content_type]},
            "course_id": query.get("course_id"),
        }
        threads = list(CommentThread().find(thread_filter))
        if ids_only:
            return [{"_id": thread["_id"]} for thread in threads]
        return threads

    @staticmethod
    def update_user(user_id: str, data: dict[str, Any]) -> int:
        """Update user."""
        return Users().update(user_id, **data)

    @staticmethod
    def get_thread_id_by_comment_id(parent_comment_id: str) -> str:
        """
        The thread Id from the parent comment.
        """
        parent_comment = Comment().get(parent_comment_id)
        if parent_comment:
            return parent_comment["comment_thread_id"]
        raise ValueError("Comment doesn't have the thread.")

    @staticmethod
    def get_course_id_by_thread_id(thread_id: str) -> str | None:
        """
        Return course_id for the matching thread.
        """
        try:
            thread = CommentThread().get(thread_id)
            if thread:
                # As thread can be a standalone thread(without a course_id).
                # So, using thread.get() instead of thread[] to avoid key error.
                return thread.get("course_id")
        except bson_errors.InvalidId:
            # this exception occurs when trying to get course_id from thread that exists
            # in mysql. Then Id will not be an ObjectID. So bypassing this exception will
            # let it search in mysql.
            pass
        return None

    @staticmethod
    def get_course_id_by_comment_id(comment_id: str) -> str | None:
        """
        Return course_id for the matching comment.
        """
        try:
            comment = Comment().get(comment_id)
            if comment:
                # As comment can be a standalone comment(comment on a standalone thread).
                # So, using comment.get() instead of comment[] to avoid key error.
                return comment.get("course_id")
        except bson_errors.InvalidId:
            # this exception occurs when trying to get course_id from comment that exists
            # in mysql. Then Id will not be an ObjectID. So bypassing this exception will
            # let it search in mysql.
            pass
        return None

    @staticmethod
    def get_users(**kwargs: Any) -> list[dict[str, Any]]:
        """Get users."""
        return list(Users().get_list(**kwargs))

    @staticmethod
    def get_user_sort_criterion(sort_by: str) -> dict[str, Any]:
        """Get sort criterion based on sort_by parameter."""
        if sort_by == "flagged":
            return {
                "course_stats.active_flags": -1,
                "course_stats.inactive_flags": -1,
                "username": -1,
            }
        elif sort_by == "recency":
            return {
                "course_stats.last_activity_at": -1,
                "username": -1,
            }
        else:
            return {
                "course_stats.threads": -1,
                "course_stats.responses": -1,
                "course_stats.replies": -1,
                "username": -1,
            }

    @staticmethod
    def create_user_pipeline(
        course_id: str, page: int, per_page: int, sort_criterion: dict[str, Any]
    ) -> list[dict[str, Any]]:
        """Get pipeline for course stats api."""
        pipeline: list[dict[str, Any]] = [
            {"$match": {"course_stats.course_id": course_id}},
            {"$project": {"username": 1, "course_stats": 1}},
            {"$unwind": "$course_stats"},
            {"$match": {"course_stats.course_id": course_id}},
            {"$sort": sort_criterion},
            {
                "$facet": {
                    "pagination": [{"$count": "total_count"}],
                    "data": [
                        {"$skip": (page - 1) * per_page},
                        {"$limit": per_page},
                    ],
                }
            },
        ]
        return pipeline

    # pylint: disable=E1121
    @classmethod
    def get_paginated_user_stats(
        cls, course_id: str, page: int, per_page: int, sort_criterion: dict[str, Any]
    ) -> dict[str, Any]:
        """Get paginated stats for a course."""
        pipeline = cls.create_user_pipeline(course_id, page, per_page, sort_criterion)
        return list(Users().aggregate(pipeline))[0]

    @staticmethod
    def get_contents(**kwargs: Any) -> list[dict[str, Any]]:
        """Return contents."""
        return list(Contents().get_list(**kwargs))

    @staticmethod
    def get_user_thread_filter(course_id: str) -> dict[str, Any]:
        """Get user thread filter."""
        return {
            "_type": {"$in": [CommentThread.content_type]},
            "course_id": {"$in": [course_id]},
        }

    @staticmethod
    def find_thread(**kwargs: Any) -> Optional[dict[str, Any]]:
        """
        Retrieves a first matching thread from the database.
        """
        return CommentThread().find_one(kwargs)

    @staticmethod
    def find_comment(
        is_parent_comment: bool = True, with_abuse_flaggers: bool = False, **kwargs: Any
    ) -> Optional[dict[str, Any]]:
        """
        Retrieves a first matching comment from the database.
        """
        if is_parent_comment:
            kwargs["parent_id"] = None
        else:
            kwargs["parent_id"] = {"$ne": None}
        if with_abuse_flaggers:
            kwargs["abuse_flaggers"] = {"$ne": []}

        return Comment().find_one(kwargs)

    @staticmethod
    def get_user_contents_by_username(username: str) -> list[dict[str, Any]]:
        """
        Retrieve all threads and comments authored by a specific user.
        """
        contents = list(Comment().find({"author_username": username})) + list(
            CommentThread().find({"author_username": username})
        )
        return contents



"""Management command for rebuild forum indices"""

from argparse import ArgumentParser

from django.core.management.base import BaseCommand

from forum.search import get_index_search_backend
from forum.utils import get_int_value_from_collection


class Command(BaseCommand):
    """
    Django management command for rebuilding search indices from scratch.
    """

    help = "Rebuild Elasticsearch indices by creating new indices and reindexing data."

    def add_arguments(self, parser: ArgumentParser) -> None:
        """
        Adds command line arguments to the rebuild_forum_indices command.

        Args:
            parser: The argument parser to which the --batch_size and --extra_catchup_minutes arguments are added.
        """
        parser.add_argument(
            "--batch_size",
            type=int,
            default=500,
            help="Number of documents to process in each batch.",
        )
        parser.add_argument(
            "--extra_catchup_minutes",
            type=int,
            default=5,
            help="Extra minutes to adjust the start time for catch-up.",
        )

    def handle(self, *args: list[str], **kwargs: dict[str, int]) -> None:
        """
        Handles the execution of the rebuild_indices command.

        Rebuilds Elasticsearch indices by creating new indices and reindexing data.
        Batch size and extra catch-up minutes can be specified.

        Args:
            args: Additional arguments.
            kwargs: Command options.
        """
        search_backend = get_index_search_backend()

        batch_size = get_int_value_from_collection(kwargs, "batch_size", 500)
        extra_catchup_minutes = get_int_value_from_collection(
            kwargs, "extra_catchup_minutes", 5
        )

        search_backend.rebuild_indices(
            batch_size=batch_size, extra_catchup_minutes=extra_catchup_minutes
        )
        self.stdout.write(self.style.SUCCESS("Forum indices rebuilt successfully."))



"""Forum Threads API Views."""

import logging
from typing import Any

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api.threads import (
    create_thread,
    delete_thread,
    get_thread,
    get_user_threads,
    update_thread,
)
from forum.utils import ForumV2RequestError, str_to_bool

log = logging.getLogger(__name__)


class ThreadsAPIView(APIView):
    """
    API view to handle operations related to threads.

    This view uses the CommentThread model for database interactions and the ThreadSerializer
    for serializing and deserializing data.
    """

    permission_classes = (AllowAny,)

    def get(self, request: Request, thread_id: str) -> Response:
        """
        Retrieve a thread by its ID.

        Args:
            request: The HTTP request object.
            thread_id: The ID of the thread to retrieve.

        Returns:
            Response: A Response object containing the serialized thread data or an error message.
        """
        try:
            params = request.query_params.dict()
            data = get_thread(thread_id, params)
        except ForumV2RequestError as error:
            return Response(
                {"error": str(error)},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response(data, status=status.HTTP_200_OK)

    def delete(self, request: Request, thread_id: str) -> Response:
        """
        Deletes a thread by its ID.

        Parameters:
            request (Request): The incoming request.
            thread_id: The ID of the thread to be deleted.
        Body:
            Empty.
        Response:
            The details of the thread that is deleted.
        """
        try:
            serialized_data = delete_thread(thread_id)
            return Response(serialized_data, status=status.HTTP_200_OK)
        except ForumV2RequestError as error:
            return Response(
                {"error": str(error)},
                status=status.HTTP_400_BAD_REQUEST,
            )

    def put(self, request: Request, thread_id: str) -> Response:
        """
        Updates an existing thread.

        Parameters:
            request (Request): The incoming request.
            thread_id: The ID of the thread to be edited.
        Body:
            fields to be updated.
        Response:
            The details of the thread that is updated.
        """

        try:
            serialized_data = update_thread(thread_id, **request.data)
            return Response(serialized_data, status=status.HTTP_200_OK)
        except ForumV2RequestError as error:
            return Response(
                {"error": str(error)},
                status=status.HTTP_400_BAD_REQUEST,
            )


class CreateThreadAPIView(APIView):
    """
    API view to create a new thread.

    This view uses the CommentThread model for database interactions and the ThreadSerializer
    for serializing and deserializing data.
    """

    permission_classes = (AllowAny,)

    def post(self, request: Request) -> Response:
        """
        Create a new thread.

        Parameters:
            request (Request): The incoming request.
        Body:
            fields to be added in a new thread.
        Response:
            The details of the thread that is created.
        """

        try:
            params = request.data
            if params.get("anonymous"):
                params["anonymous"] = str_to_bool(params["anonymous"])
            if params.get("anonymous_to_peers"):
                params["anonymous_to_peers"] = str_to_bool(params["anonymous_to_peers"])
            serialized_data = create_thread(**params)
            return Response(serialized_data, status=status.HTTP_200_OK)
        except (TypeError, ForumV2RequestError) as error:
            return Response(
                {"error": str(error)},
                status=status.HTTP_400_BAD_REQUEST,
            )


class UserThreadsAPIView(APIView):
    """
    API View for getting all threads of a course.

    This view provides an endpoint for retrieving all threads based on course id.
    """

    permission_classes = (AllowAny,)

    def get(self, request: Request) -> Response:
        """
        Retrieve a course's threads.

        Args:
            request (HttpRequest): The HTTP request object.

        Returns:
            Response: A Response object with the threads data.

        Raises:
            HTTP_400_BAD_REQUEST: If the user does not exist.
        """
        try:
            params: dict[str, Any] = request.GET.dict()
            serialized_data = get_user_threads(**params)
            return Response(serialized_data, status=status.HTTP_200_OK)
        except (TypeError, ValueError, ForumV2RequestError) as error:
            return Response(
                {"error": str(error)},
                status=status.HTTP_400_BAD_REQUEST,
            )



"""Management command for creating mongodb indexes"""

from django.core.management.base import BaseCommand

from forum.backends.mongodb import BaseContents


class Command(BaseCommand):
    """
    Django management command for creating mongodb indexes.
    """

    help = "Create or Update indexes in the mongodb for the content model"

    def handle(self, *args: list[str], **kwargs: dict[str, str]) -> None:
        """
        Handles the execution of the forum_create_mongodb_indexes command.

        This command creates or updates indexes in the mongodb for the content model.
        """
        BaseContents().create_indexes()
        self.stdout.write(
            self.style.SUCCESS("Created/Updated Mongodb indexes successfuly.")
        )



"""Mongo module for forum app."""

import logging
from typing import Any, Optional

from django.conf import settings
from pymongo import MongoClient
from pymongo.database import Database as PymongoDatabase

log = logging.getLogger(__name__)

Database = PymongoDatabase[dict[str, Any]]


def get_database(
    database: Optional[str] = None,
    client_params: Optional[dict[Any, Any]] = None,
) -> Database:
    """
    Connect to MongoDB.

    :Parameters:

      - `database`: name of the mongodb database
      - `client_params`: parameters passed to MongoClient(...)
    """
    if database is None:
        database = settings.FORUM_MONGODB_DATABASE
    if client_params is None:
        client_params = settings.FORUM_MONGODB_CLIENT_PARAMETERS

    connection: MongoClient[Any] = MongoClient(**client_params)
    db = connection[database]
    return db



"""
Native Python Comments APIs.
"""

import logging
import math
from typing import Any, Optional

from django.core.exceptions import ObjectDoesNotExist
from rest_framework.serializers import ValidationError

from forum.backend import get_backend
from forum.serializers.comment import CommentSerializer
from forum.utils import ForumV2RequestError

log = logging.getLogger(__name__)


def prepare_comment_api_response(
    comment: dict[str, Any],
    backend: Any,
    exclude_fields: Optional[list[str]] = None,
) -> dict[str, Any]:
    """
    Return serialized validated data.

    Parameters:
        comment: The comment details that needs to be serialized.
        exclude_fields: Any fields that need to be excluded from response.

    Response:
        serialized validated data of the comment.
    """
    comment_data = {
        **comment,
        "id": str(comment.get("_id")),
        "user_id": comment.get("author_id"),
        "thread_id": str(comment.get("comment_thread_id")),
        "username": comment.get("author_username"),
        "parent_id": str(comment.get("parent_id")),
        "type": str(comment.get("_type", "")).lower(),
    }
    if not exclude_fields:
        exclude_fields = []
    exclude_fields.append("children")
    serializer = CommentSerializer(
        data=comment_data,
        exclude_fields=exclude_fields,
        backend=backend,
    )
    if not serializer.is_valid(raise_exception=True):
        raise ValidationError(serializer.errors)

    return serializer.data


def get_parent_comment(
    comment_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Get a parent comment.

    Parameters:
        comment_id: The ID of the comment.
    Body:
        Empty.
    Response:
        The details of the comment for the given comment_id.
    """
    backend = get_backend(course_id)()
    try:
        comment = backend.validate_object("Comment", comment_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for get parent comment request.")
        raise ForumV2RequestError(
            f"Comment does not exists with Id: {comment_id}"
        ) from exc
    return prepare_comment_api_response(
        comment,
        backend,
        exclude_fields=["sk"],
    )


def create_child_comment(
    parent_comment_id: str,
    body: str,
    user_id: str,
    course_id: str,
    anonymous: bool,
    anonymous_to_peers: bool,
) -> dict[str, Any]:
    """
    Create a new child comment.

    Parameters:
        comment_id: The ID of the parent comment for creating it's child comment.
        body: The content of the comment.
        course_id: The Id of the respective course.
        user_id: The requesting user id.
        anonymous: anonymous flag(True or False).
        anonymous_to_peers: anonymous to peers flag(True or False).
    Response:
        The details of the comment that is created.
    """
    backend = get_backend(course_id)()
    try:
        parent_comment = backend.validate_object("Comment", parent_comment_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for create child comment request.")
        raise ForumV2RequestError(
            f"Comment does not exists with Id: {parent_comment_id}"
        ) from exc

    comment_id = backend.create_comment(
        {
            "body": body,
            "author_id": user_id,
            "course_id": course_id,
            "anonymous": anonymous,
            "anonymous_to_peers": anonymous_to_peers,
            "depth": 1,
            "comment_thread_id": backend.get_thread_id_by_comment_id(parent_comment_id),
            "parent_id": parent_comment_id,
        }
    )
    comment = backend.get_comment(comment_id)
    if not comment:
        log.error("Forumv2RequestError for create child comment request.")
        raise ForumV2RequestError("comment is not created")

    user = backend.get_user(user_id)
    thread = backend.get_thread(parent_comment["comment_thread_id"])
    if user and thread and comment:
        backend.mark_as_read(user_id, parent_comment["comment_thread_id"])
    try:
        comment_data = prepare_comment_api_response(
            comment,
            backend,
            exclude_fields=["endorsement", "sk"],
        )
        return comment_data
    except ValidationError as error:
        raise error


def update_comment(
    comment_id: str,
    body: Optional[str] = None,
    course_id: Optional[str] = None,
    user_id: Optional[str] = None,
    anonymous: Optional[bool] = None,
    anonymous_to_peers: Optional[bool] = None,
    endorsed: Optional[bool] = None,
    closed: Optional[bool] = None,
    editing_user_id: Optional[str] = None,
    edit_reason_code: Optional[str] = None,
    endorsement_user_id: Optional[str] = None,
) -> dict[str, Any]:
    """
    Update an existing child/parent comment.

    Parameters:
        comment_id: The ID of the comment to be edited.
        body (Optional[str]): The content of the comment.
        course_id (Optional[str]): The Id of the respective course.
        user_id (Optional[str]): The requesting user id.
        anonymous (Optional[bool]): anonymous flag(True or False).
        anonymous_to_peers (Optional[bool]): anonymous to peers flag(True or False).
        endorsed (Optional[bool]): Flag indicating if the comment is endorsed by any user.
        closed (Optional[bool]): Flag indicating if the comment thread is closed.
        editing_user_id (Optional[str]): The ID of the user editing the comment.
        edit_reason_code (Optional[str]): The reason for editing the comment, typically represented by a code.
        endorsement_user_id (Optional[str]): The ID of the user endorsing the comment.
    Response:
        The details of the comment that is updated.
    """
    backend = get_backend(course_id)()
    try:
        backend.validate_object("Comment", comment_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for update comment request.")
        raise ForumV2RequestError(
            f"Comment does not exists with Id: {comment_id}"
        ) from exc

    updated_comment = backend.update_comment_and_get_updated_comment(
        comment_id,
        body=body,
        user_id=user_id,
        anonymous=anonymous,
        anonymous_to_peers=anonymous_to_peers,
        endorsed=endorsed,
        closed=closed,
        editing_user_id=editing_user_id,
        edit_reason_code=edit_reason_code,
        endorsement_user_id=endorsement_user_id,
    )
    if not updated_comment:
        log.error("Forumv2RequestError for create child comment request.")
        raise ForumV2RequestError("comment is not updated")
    try:
        return prepare_comment_api_response(
            updated_comment,
            backend,
            exclude_fields=(
                ["endorsement", "sk"] if updated_comment.get("parent_id") else ["sk"]
            ),
        )
    except ValidationError as error:
        raise error


def delete_comment(comment_id: str, course_id: Optional[str] = None) -> dict[str, Any]:
    """
    Delete a comment.

    Parameters:
        comment_id: The ID of the comment to be deleted.
    Body:
        Empty.
    Response:
        The details of the comment that is deleted.
    """
    backend = get_backend(course_id)()
    try:
        comment = backend.validate_object("Comment", comment_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for delete comment request.")
        raise ForumV2RequestError(
            f"Comment does not exists with Id: {comment_id}"
        ) from exc
    data = prepare_comment_api_response(
        comment,
        backend,
        exclude_fields=["endorsement", "sk"],
    )
    backend.delete_comment(comment_id)
    author_id = comment["author_id"]
    comment_course_id = comment["course_id"]
    parent_comment_id = data["parent_id"]
    if parent_comment_id:
        backend.update_stats_for_course(author_id, comment_course_id, replies=-1)
    else:
        backend.update_stats_for_course(author_id, comment_course_id, responses=-1)
    return data


def create_parent_comment(
    thread_id: str,
    body: str,
    user_id: str,
    course_id: str,
    anonymous: bool,
    anonymous_to_peers: bool,
) -> dict[str, Any]:
    """
    Create a new parent comment.

    Parameters:
        thread_id: The ID of the thread for creating a comment on it.
        body: The content of the comment.
        course_id: The Id of the respective course.
        user_id: The requesting user id.
        anonymous: anonymous flag(True or False).
        anonymous_to_peers: anonymous to peers flag(True or False).
    Response:
        The details of the comment that is created.
    """
    backend = get_backend(course_id)()
    try:
        backend.validate_object("CommentThread", thread_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for create parent comment request.")
        raise ForumV2RequestError(
            f"Thread does not exists with Id: {thread_id}"
        ) from exc

    comment_id = backend.create_comment(
        {
            "body": body,
            "author_id": user_id,
            "course_id": course_id,
            "anonymous": anonymous,
            "anonymous_to_peers": anonymous_to_peers,
            "depth": 0,
            "comment_thread_id": thread_id,
        }
    )
    if not comment_id:
        log.error("Forumv2RequestError for create parent comment request.")
        raise ForumV2RequestError("comment is not created")
    comment = backend.get_comment(comment_id) or {}
    user = backend.get_user(user_id)
    if user and comment:
        backend.mark_as_read(user_id, thread_id)
    try:
        return prepare_comment_api_response(
            comment,
            backend,
            exclude_fields=["endorsement", "sk"],
        )
    except ValidationError as error:
        raise error


def get_course_id_by_comment(comment_id: str) -> str | None:
    """
    Return course_id for the matching comment.
    It searches for comment_id both in mongodb and mysql.
    """
    #  pylint: disable=C0415
    from forum.backends.mongodb.api import MongoBackend
    from forum.backends.mysql.api import MySQLBackend

    return (
        MongoBackend.get_course_id_by_comment_id(comment_id)
        or MySQLBackend.get_course_id_by_comment_id(comment_id)
        or None
    )


def get_user_comments(
    user_id: str,
    course_id: str,
    flagged: Optional[bool] = False,
    page: int = 1,
    per_page: int = 10,
) -> dict[str, Any]:
    """
    Get all comments made by a user in a specific course.

    Args:
        user_id: The ID of the user
        course_id: The ID of the course
        flagged: Filter for flagged comments
        page: Page number for pagination
        per_page: Number of items per page

    Returns:
        A dictionary containing paginated comment results
    """

    backend = get_backend(course_id)()

    # Build query parameters
    query_params: dict[str, Any] = {
        "author_id": str(user_id),
        "course_id": course_id,
    }

    if flagged:
        query_params["abuse_flaggers"] = {"$ne": [], "$exists": True}

    # Get total count
    comment_count = backend.get_comments_count(**query_params)

    # Calculate pagination
    num_pages = max(1, math.ceil(comment_count / per_page))
    skip = (page - 1) * per_page

    # Get paginated comments
    query_params["resp_skip"] = skip
    query_params["resp_limit"] = per_page
    query_params["sort"] = -1  # Sort by newest first

    comments = backend.get_comments(**query_params)

    return {
        "collection": comments,
        "comment_count": comment_count,
        "num_pages": num_pages,
        "page": page,
    }



"""Admin module for forum."""

from django.contrib import admin
from forum.models import (
    ForumUser,
    CourseStat,
    CommentThread,
    Comment,
    EditHistory,
    AbuseFlagger,
    HistoricalAbuseFlagger,
    ReadState,
    LastReadTime,
    UserVote,
    Subscription,
    MongoContent,
)


@admin.register(ForumUser)
class ForumUserAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for ForumUser model."""

    list_display = ("user", "default_sort_key")
    search_fields = ("user__username", "user__email")


@admin.register(CourseStat)
class CourseStatAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for CourseStat model."""

    list_display = (
        "user",
        "course_id",
        "active_flags",
        "inactive_flags",
        "threads",
        "responses",
        "replies",
        "last_activity_at",
    )
    search_fields = ("user__username", "course_id")
    list_filter = ("course_id",)


@admin.register(CommentThread)
class CommentThreadAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for CommentThread model."""

    list_display = (
        "title",
        "author",
        "course_id",
        "thread_type",
        "context",
        "closed",
        "pinned",
        "created_at",
        "updated_at",
    )
    search_fields = ("title", "body", "author__username", "course_id")
    list_filter = ("thread_type", "context", "closed", "pinned")


@admin.register(Comment)
class CommentAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for Comment model."""

    list_display = (
        "comment_thread",
        "author",
        "body",
        "created_at",
        "updated_at",
        "endorsed",
        "anonymous",
    )
    search_fields = ("body", "author__username", "comment_thread__title")
    list_filter = ("endorsed", "anonymous")


@admin.register(EditHistory)
class EditHistoryAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for EditHistory model."""

    list_display = ("editor", "content_object_id", "created_at", "reason_code")
    search_fields = ("editor__username", "original_body")
    list_filter = ("reason_code",)


@admin.register(AbuseFlagger)
class AbuseFlaggerAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for AbuseFlagger model."""

    list_display = ("user", "content_object_id", "flagged_at")
    search_fields = ("user__username",)
    list_filter = ("content_type",)


@admin.register(HistoricalAbuseFlagger)
class HistoricalAbuseFlaggerAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for HistoricalAbuseFlagger model."""

    list_display = ("user", "content_object_id", "flagged_at")
    search_fields = ("user__username",)
    list_filter = ("content_type",)


@admin.register(ReadState)
class ReadStateAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for ReadState model."""

    list_display = ("user", "course_id")
    search_fields = ("user__username", "course_id")


@admin.register(LastReadTime)
class LastReadTimeAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for LastReadTime model."""

    list_display = ("read_state", "comment_thread", "timestamp")
    search_fields = ("read_state__user__username", "comment_thread__title")


@admin.register(UserVote)
class UserVoteAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for UserVote model."""

    list_display = ("user", "content_object_id", "vote")
    search_fields = ("user__username",)
    list_filter = ("vote",)


@admin.register(Subscription)
class SubscriptionAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for Subscription model."""

    list_display = (
        "subscriber",
        "source_object_id",
        "source_content_type",
        "created_at",
        "updated_at",
    )
    search_fields = ("subscriber__username",)
    list_filter = ("source_content_type",)


@admin.register(MongoContent)
class MongoContentAdmin(admin.ModelAdmin):  # type: ignore
    """Admin interface for MongoContent model."""

    list_display = ("mongo_id", "content_object_id", "content_type")
    search_fields = ("mongo_id",)



"""
API for subscriptions.
"""

from typing import Any, Optional

from django.http import QueryDict
from rest_framework.request import Request
from rest_framework.test import APIRequestFactory

from forum.backend import get_backend
from forum.pagination import ForumPagination
from forum.serializers.subscriptions import SubscriptionSerializer
from forum.serializers.thread import ThreadSerializer
from forum.utils import ForumV2RequestError
from forum.constants import FORUM_DEFAULT_PAGE, FORUM_DEFAULT_PER_PAGE


def validate_user_and_thread(
    user_id: str, source_id: str, course_id: Optional[str] = None
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Validate if user and thread exist.
    """
    backend = get_backend(course_id)()
    user = backend.get_user(user_id)
    thread = backend.get_thread(source_id)
    if not (user and thread):
        raise ForumV2RequestError("User / Thread doesn't exist")
    return user, thread


def create_subscription(
    user_id: str, source_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Create a subscription for a user.
    """
    backend = get_backend(course_id)()
    _, _ = validate_user_and_thread(user_id, source_id, course_id=course_id)
    subscription = backend.subscribe_user(
        user_id, source_id, source_type="CommentThread"
    )
    serializer = SubscriptionSerializer(subscription)
    return serializer.data


def delete_subscription(
    user_id: str, source_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Delete a subscription for a user.
    """
    backend = get_backend(course_id)()
    _, _ = validate_user_and_thread(user_id, source_id, course_id=course_id)

    subscription = backend.get_subscription(
        user_id, source_id, source_type="CommentThread"
    )
    if not subscription:
        raise ForumV2RequestError("Subscription doesn't exist")

    backend.unsubscribe_user(user_id, source_id, source_type="CommentThread")
    serializer = SubscriptionSerializer(subscription)
    return serializer.data


def get_user_subscriptions(
    user_id: str,
    course_id: str,
    author_id: Optional[str] = None,
    thread_type: Optional[str] = None,
    flagged: Optional[bool] = False,
    unread: Optional[bool] = False,
    unanswered: Optional[bool] = False,
    unresponded: Optional[bool] = False,
    count_flagged: Optional[bool] = False,
    sort_key: Optional[str] = "date",
    page: Optional[int] = FORUM_DEFAULT_PAGE,
    per_page: Optional[int] = FORUM_DEFAULT_PER_PAGE,
    group_id: Optional[str] = None,
    group_ids: Optional[str] = None,
) -> dict[str, Any]:
    """
    Get a user's subscriptions.
    """
    backend = get_backend(course_id)()
    params = {
        "course_id": course_id,
        "author_id": author_id,
        "thread_type": thread_type,
        "flagged": flagged,
        "unread": unread,
        "unanswered": unanswered,
        "unresponded": unresponded,
        "count_flagged": count_flagged,
        "sort_key": sort_key,
        "page": int(page) if page else None,
        "per_page": int(per_page) if per_page else None,
        "user_id": user_id,
        "group_id": group_id,
        "group_ids": group_ids,
    }
    params = {k: v for k, v in params.items() if v is not None}
    thread_ids = backend.find_subscribed_threads(user_id, course_id)
    threads = backend.get_threads(params, user_id, ThreadSerializer, thread_ids)
    return threads


def get_thread_subscriptions(
    thread_id: str, page: int = 1, per_page: int = 20, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Retrieve subscriptions to a specific thread.

    Args:
        thread_id (str): The ID of the thread to retrieve subscriptions for.
        page (int): The page number for pagination.
        per_page (int): The number of items per page.

    Returns:
        dict: A dictionary containing the paginated subscription data.
    """
    backend = get_backend(course_id)()
    query = {"source_id": thread_id, "source_type": "CommentThread"}
    subscriptions_list = list(backend.get_subscriptions(query))

    factory = APIRequestFactory()
    query_params = QueryDict("", mutable=True)
    query_params.update({"page": str(page), "per_page": str(per_page)})
    request = factory.get("/", query_params)
    drf_request = Request(request)

    paginator = ForumPagination()
    paginated_subscriptions = paginator.paginate_queryset(
        subscriptions_list, drf_request
    )

    subscriptions = SubscriptionSerializer(paginated_subscriptions, many=True)
    subscriptions_count = len(subscriptions.data)

    return {
        "collection": subscriptions.data,
        "subscriptions_count": subscriptions_count,
        "page": page,
        "num_pages": max(1, subscriptions_count // per_page),
    }



"""Deletion command for a course data in mongodb."""

from typing import Any

from django.core.management.base import BaseCommand, CommandParser

from forum.migration_helpers import delete_course_data, get_all_course_ids
from forum.mongo import get_database


class Command(BaseCommand):
    """Deletion command for a course data in mongodb."""

    help = "Delete data from MongoDB for specific courses or all courses"

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "courses", nargs="+", type=str, help="List of course IDs or `all`"
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Perform a dry run without actually deleting data",
        )

    def handle(self, *args: str, **options: dict[str, Any]) -> None:
        """Handle method for command."""
        db = get_database()

        courses: list[str] = list(options["courses"])
        dry_run: bool = bool(options["dry_run"])

        if dry_run:
            self.stdout.write(
                self.style.WARNING("Performing dry run. No data will be deleted.")
            )

        if "all" in courses:
            courses = get_all_course_ids(db)

        for course_id in courses:
            self.stdout.write(f"Deleting data for course: {course_id}")
            delete_course_data(db, course_id, dry_run, self.stdout)

        if dry_run:
            self.stdout.write(
                self.style.SUCCESS("Dry run completed. No data was deleted.")
            )
        else:
            self.stdout.write(
                self.style.SUCCESS("Data deletion completed successfully")
            )



"""
Search backend API.
"""

import importlib

from django.conf import settings

from forum.search import base


def get_index_search_backend() -> base.BaseIndexSearchBackend:
    """
    Return backend instance for managing search indices.
    """
    return _get_search_backend().INDEX_SEARCH_CLASS()


def get_document_search_backend() -> base.BaseDocumentSearchBackend:
    """
    Return backend instance for indexing, deleting and updating documents.
    """
    return _get_search_backend().DOCUMENT_SEARCH_CLASS()


def get_thread_search_backend() -> base.BaseThreadSearchBackend:
    """
    Return backend instance for searching thread documents.
    """
    return _get_search_backend().THREAD_SEARCH_CLASS()


def _get_search_backend() -> base.BaseSearchBackend:
    """
    Future-proof backend provider that is compatible with multiple
    backends.

    By default, use the Elasticsearch search backend.
    """
    search_backend_module_name = getattr(
        settings, "FORUM_SEARCH_BACKEND", "forum.search.es.ElasticsearchBackend"
    )
    module_name, class_name = search_backend_module_name.rsplit(".", 1)
    Backend = getattr(importlib.import_module(module_name), class_name)
    return Backend



"""Subscriptions class for mongo backend."""

from datetime import datetime, timezone
from typing import Any, Optional

from forum.backends.mongodb.base_model import MongoBaseModel


class Subscriptions(MongoBaseModel):
    """
    Represents a subscription to a source in the MongoDB database.

    This class provides methods for inserting, updating, retrieving, and listing subscriptions.
    """

    COLLECTION_NAME: str = "subscriptions"

    def insert(self, subscriber_id: str, source_id: str, source_type: str) -> str:
        """
        Inserts a new subscription into the MongoDB collection.

        Args:
            subscriber_id: The ID of the subscriber.
            source_id: The ID of the source.
            source_type: The type of the source.

        """
        subscription = {
            "subscriber_id": subscriber_id,
            "source_id": source_id,
            "source_type": source_type,
            "created_at": datetime.now(timezone.utc),
            "updated_at": datetime.now(timezone.utc),
        }
        result = self._collection.insert_one(subscription)
        return str(result.inserted_id)

    def update(self, subscriber_id: str, source_id: str, **kwargs: Any) -> int:
        """
        Updates an existing subscription in the MongoDB collection.

        Args:
            subscriber_id: The ID of the subscriber.
            source_id: The ID of the source.
            **kwargs: Additional fields to update.

        """
        filter_query = {
            "subscriber_id": subscriber_id,
            "source_id": source_id,
        }
        update_query = {"$set": kwargs, "$currentDate": {"updated_at": True}}
        result = self._collection.update_one(filter_query, update_query)
        return result.modified_count

    def get_subscription(
        self, subscriber_id: str, source_id: str
    ) -> Optional[dict[str, Any]]:
        """
        Retrieves a subscription from the MongoDB collection.

        Args:
            subscriber_id: The ID of the subscriber.
            source_id: The ID of the source.

        Returns:
            The subscription document if found, otherwise None.

        """
        filter_query = {
            "subscriber_id": subscriber_id,
            "source_id": source_id,
        }
        subscription = self._collection.find_one(filter_query)
        return subscription

    def delete_subscription(
        self, subscriber_id: str, source_id: str, source_type: Optional[str] = ""
    ) -> int:
        """
        Deletes a subscription from the MongoDB collection.

        Args:
            subscriber_id: The ID of the subscriber.
            source_id: The ID of the source.

        Returns:
            The number of deleted documents.

        """
        filter_query = {
            "subscriber_id": subscriber_id,
            "source_id": source_id,
        }
        if source_type:
            filter_query["source_type"] = source_type

        result = self._collection.delete_one(filter_query)
        return result.deleted_count



"""
Constants values used in the Forum App.
"""

FORUM_DEFAULT_PAGE = 1
FORUM_DEFAULT_PER_PAGE = 20

FORUM_MAX_DEEP_SEARCH_COMMENT_COUNT = 1000

RETIRED_TITLE = "[deleted]"
RETIRED_BODY = "[deleted]"



"""
Native Python Users APIs.
"""

import logging
import math
from typing import Any, Optional

from forum.backend import get_backend
from forum.constants import FORUM_DEFAULT_PAGE, FORUM_DEFAULT_PER_PAGE
from forum.serializers.thread import ThreadSerializer
from forum.serializers.users import UserSerializer
from forum.utils import ForumV2RequestError

log = logging.getLogger(__name__)


def get_user(
    user_id: str,
    group_ids: Optional[list[int]] = None,
    course_id: Optional[str] = None,
    complete: Optional[bool] = False,
) -> dict[str, Any]:
    """Get user data by user_id."""
    """
    Get users data by user_id.
    Parameters:
        user_id (str): The ID of the requested User.
        params (str): attributes for user's data filteration.
    Response:
        A response with the users data.
    """
    backend = get_backend(course_id)()
    user = backend.get_user(user_id, get_full_dict=False)
    if not user:
        log.error(f"Forumv2RequestError for retrieving user's data for id {user_id}.")
        raise ForumV2RequestError(str(f"user not found with id: {user_id}"))

    params = {
        "complete": complete,
        "group_ids": group_ids,
        "course_id": course_id,
    }
    hashed_user = backend.user_to_hash(user_id, params)
    serializer = UserSerializer(hashed_user)
    return serializer.data


def update_user(
    user_id: str,
    username: Optional[str] = None,
    default_sort_key: Optional[str] = None,
    course_id: Optional[str] = None,
    group_ids: Optional[list[int]] = None,
    complete: Optional[bool] = False,
) -> dict[str, Any]:
    """Update user."""
    backend = get_backend(course_id)()
    user = backend.get_user(user_id)
    user_by_username = backend.get_user_by_username(username)
    if user and user_by_username:
        if user["external_id"] != user_by_username["external_id"]:
            raise ForumV2RequestError("user does not match")
    elif user_by_username:
        raise ForumV2RequestError(f"user already exists with username: {username}")
    else:
        user_id = backend.find_or_create_user(user_id)
    update_data = {"username": username}
    if default_sort_key is not None:
        update_data["default_sort_key"] = default_sort_key
    backend.update_user(user_id, update_data)
    updated_user = backend.get_user(user_id)
    if not updated_user:
        raise ForumV2RequestError(f"user not found with id: {user_id}")
    params = {
        "complete": complete,
        "group_ids": group_ids,
        "course_id": course_id,
    }
    hashed_user = backend.user_to_hash(user_id, params)
    serializer = UserSerializer(hashed_user)
    return serializer.data


def create_user(
    user_id: str,
    username: str,
    default_sort_key: str = "date",
    course_id: Optional[str] = None,
    group_ids: Optional[list[int]] = None,
    complete: bool = False,
) -> dict[str, Any]:
    """Create user."""
    backend = get_backend(course_id)()
    user_by_id = backend.get_user(user_id)
    user_by_username = backend.get_user_by_username(username)

    if user_by_id or user_by_username:
        raise ForumV2RequestError(f"user already exists with id: {id}")

    backend.find_or_create_user(
        user_id, username=username, default_sort_key=default_sort_key
    )
    user = backend.get_user(user_id)
    if not user:
        raise ForumV2RequestError(f"user not found with id: {user_id}")
    params = {
        "complete": complete,
        "group_ids": group_ids,
        "course_id": course_id,
    }
    hashed_user = backend.user_to_hash(user_id, params)
    serializer = UserSerializer(hashed_user)
    return serializer.data


def update_username(
    user_id: str, new_username: str, course_id: Optional[str] = None
) -> dict[str, str]:
    """Update username."""
    backend = get_backend(course_id)()
    user = backend.get_user(user_id)
    if not user:
        raise ForumV2RequestError(str(f"user not found with id: {user_id}"))
    backend.update_user(user_id, {"username": new_username})
    backend.replace_username_in_all_content(user_id, new_username)
    return {"message": "Username updated successfully"}


def retire_user(
    user_id: str, retired_username: str, course_id: Optional[str] = None
) -> dict[str, str]:
    """Retire user."""
    backend = get_backend(course_id)()
    user = backend.get_user(user_id)
    if not user:
        raise ForumV2RequestError(f"user not found with id: {user_id}")
    backend.update_user(
        user_id,
        data={
            "email": "",
            "username": retired_username,
            "read_states": [],
        },
    )
    backend.unsubscribe_all(user_id)
    backend.retire_all_content(user_id, retired_username)

    return {"message": "User retired successfully"}


def mark_thread_as_read(
    user_id: str,
    source_id: str,
    complete: bool = False,
    course_id: Optional[str] = None,
    group_ids: Optional[list[int]] = None,
) -> dict[str, Any]:
    """Mark thread as read."""
    backend = get_backend(course_id)()
    user = backend.get_user(user_id)
    if not user:
        raise ForumV2RequestError(str(f"user not found with id: {user_id}"))

    thread = backend.get_thread(source_id)
    if not thread:
        raise ForumV2RequestError(str(f"source not found with id: {source_id}"))

    backend.mark_as_read(user_id, source_id)

    user = backend.get_user(user_id)
    if not user:
        raise ForumV2RequestError(str(f"user not found with id: {user_id}"))

    params = {
        "complete": complete,
        "group_ids": group_ids,
        "course_id": course_id,
    }

    hashed_user = backend.user_to_hash(user_id, params)
    serializer = UserSerializer(hashed_user)
    return serializer.data


def get_user_active_threads(
    user_id: str,
    course_id: str,
    author_id: Optional[str] = None,
    thread_type: Optional[str] = None,
    flagged: Optional[bool] = False,
    unread: Optional[bool] = False,
    unanswered: Optional[bool] = False,
    unresponded: Optional[bool] = False,
    count_flagged: Optional[bool] = False,
    sort_key: Optional[str] = "user_activity",
    page: Optional[int] = FORUM_DEFAULT_PAGE,
    per_page: Optional[int] = FORUM_DEFAULT_PER_PAGE,
    group_id: Optional[str] = None,
    is_moderator: Optional[bool] = False,
) -> dict[str, Any]:
    """Get user active threads."""
    backend = get_backend(course_id)()
    raw_query = bool(sort_key == "user_activity")
    if not course_id:
        return {}
    active_contents = list(
        backend.get_contents(
            author_id=user_id,
            anonymous=False,
            anonymous_to_peers=False,
            course_id=course_id,
        )
    )

    if flagged:
        active_contents = [
            content
            for content in active_contents
            if content["abuse_flaggers"] and len(content["abuse_flaggers"]) > 0
        ]
    active_contents = sorted(
        active_contents, key=lambda x: x["updated_at"], reverse=True
    )
    active_thread_ids = list(
        set(
            (
                content["comment_thread_id"]
                if content["_type"] == "Comment"
                else content["_id"]
            )
            for content in active_contents
        )
    )

    params: dict[str, Any] = {
        "comment_thread_ids": active_thread_ids,
        "user_id": user_id,
        "course_id": course_id,
        "group_ids": [int(group_id)] if group_id else [],
        "author_id": author_id,
        "thread_type": thread_type,
        "filter_flagged": flagged,
        "filter_unread": unread,
        "filter_unanswered": unanswered,
        "filter_unresponded": unresponded,
        "count_flagged": count_flagged,
        "sort_key": sort_key,
        "page": page,
        "per_page": per_page,
        "context": "course",
        "raw_query": raw_query,
        "is_moderator": is_moderator,
    }
    data = backend.handle_threads_query(**params)

    if collections := data.get("collection"):
        thread_serializer = ThreadSerializer(
            collections,
            many=True,
            context={
                "count_flagged": count_flagged,
                "include_endorsed": True,
                "include_read_state": True,
            },
            backend=backend,
        )
        data["collection"] = thread_serializer.data
    else:
        collection = data.get("result", [])
        for thread in collection:
            thread["_id"] = str(thread.pop("_id"))
            thread["type"] = str(thread.get("_type", "")).lower()
        data["collection"] = ThreadSerializer(
            collection, many=True, backend=backend
        ).data

    return data


def _get_user_data(
    user_stats: dict[str, Any], exclude_from_stats: list[str]
) -> dict[str, Any]:
    """Get user data from user stats."""
    user_data = {"username": user_stats["username"]}
    for k, v in user_stats["course_stats"].items():
        if k not in exclude_from_stats:
            user_data[k] = v
    return user_data


def _get_stats_for_usernames(
    course_id: str, usernames: list[str], backend: Any
) -> list[dict[str, Any]]:
    """Get stats for specific usernames."""
    users = backend.get_users()
    stats_query = []
    for user in users:
        if user["username"] not in usernames:
            continue
        course_stats = user.get("course_stats")
        if course_stats:
            for course_stat in course_stats:
                if course_stat["course_id"] == course_id:
                    stats_query.append(
                        {"username": user["username"], "course_stats": course_stat}
                    )
                    break
    return sorted(stats_query, key=lambda u: usernames.index(u["username"]))


def get_user_course_stats(
    course_id: str,
    usernames: Optional[str] = None,
    page: int = FORUM_DEFAULT_PAGE,
    per_page: int = FORUM_DEFAULT_PER_PAGE,
    sort_key: str = "",
    with_timestamps: bool = False,
) -> dict[str, Any]:
    """Get user course stats."""
    backend = get_backend(course_id)()
    sort_criterion = backend.get_user_sort_criterion(sort_key)
    exclude_from_stats = ["_id", "course_id"]
    if not with_timestamps:
        exclude_from_stats.append("last_activity_at")

    usernames_list = usernames.split(",") if usernames else None
    data = []

    if not usernames_list:
        paginated_stats = backend.get_paginated_user_stats(
            course_id, page, per_page, sort_criterion
        )
        num_pages = 0
        page = 0
        total_count = 0
        if paginated_stats.get("pagination"):
            total_count = paginated_stats["pagination"][0]["total_count"]
            num_pages = max(1, math.ceil(total_count / per_page))
            data = [
                _get_user_data(user_stats, exclude_from_stats)
                for user_stats in paginated_stats["data"]
            ]
    else:
        stats_query = _get_stats_for_usernames(course_id, usernames_list, backend)
        total_count = len(stats_query)
        num_pages = 1
        data = [
            {
                "username": user_stats["username"],
                **{
                    k: v
                    for k, v in user_stats["course_stats"].items()
                    if k not in exclude_from_stats
                },
            }
            for user_stats in stats_query
        ]

    return {
        "user_stats": data,
        "num_pages": num_pages,
        "page": page,
        "count": total_count,
    }


def update_users_in_course(course_id: str) -> dict[str, int]:
    """Update all user stats in a course."""
    backend = get_backend(course_id)()
    updated_users = backend.update_all_users_in_course(course_id)
    return {"user_count": len(updated_users)}



"""Comment Class for mongo backend."""

from datetime import datetime
from typing import Any, Optional

from bson import ObjectId

from forum.backends.mongodb.contents import BaseContents
from forum.backends.mongodb.threads import CommentThread
from forum.backends.mongodb.users import Users
from forum.utils import get_handler_by_name


class Comment(BaseContents):
    """
    Comment class for cs_comments_service content model
    """

    index_name = "comments"
    content_type = "Comment"

    def override_query(self, query: dict[str, Any]) -> dict[str, Any]:
        query = {**query, "_type": self.content_type}
        return super().override_query(query)

    @classmethod
    def mapping(cls) -> dict[str, Any]:
        """
        Mapping function for the Thread class
        """
        return {
            "dynamic": "false",
            "properties": {
                "body": {
                    "type": "text",
                    "store": True,
                    "term_vector": "with_positions_offsets",
                },
                "course_id": {"type": "keyword"},
                "comment_thread_id": {"type": "keyword"},
                "commentable_id": {"type": "keyword"},
                "group_id": {"type": "keyword"},
                "context": {"type": "keyword"},
                "created_at": {"type": "date"},
                "updated_at": {"type": "date"},
                "title": {"type": "keyword"},
            },
        }

    @classmethod
    def doc_to_hash(cls, doc: dict[str, Any]) -> dict[str, Any]:
        """
        Converts comment document to the dict
        """
        return {
            "body": doc.get("body"),
            "course_id": doc.get("course_id"),
            "comment_thread_id": str(doc.get("comment_thread_id")),
            "commentable_id": doc.get("commentable_id"),
            "group_id": doc.get("group_id"),
            "context": doc.get("context", "course"),
            "created_at": doc.get("created_at"),
            "updated_at": doc.get("updated_at"),
            "title": doc.get("title"),
        }

    def insert(
        self,
        body: str,
        course_id: str,
        author_id: str,
        comment_thread_id: str,
        parent_id: Optional[str] = None,
        author_username: Optional[str] = None,
        anonymous: bool = False,
        anonymous_to_peers: bool = False,
        depth: int = 0,
        abuse_flaggers: Optional[list[str]] = None,
        historical_abuse_flaggers: Optional[list[str]] = None,
        visible: bool = True,
    ) -> str:
        """
        Inserts a new comment document into the database.

        Args:
            body (str): The body content of the comment.
            course_id (str): The ID of the course the comment is associated with.
            comment_thread_id (str): The ID of the parent comment thread.
            author_id (str): The ID of the author who created the comment.
            author_username (str): The username of the author.
            anonymous (bool, optional): Whether the comment is posted anonymously. Defaults to False.
            anonymous_to_peers (bool, optional): Whether the comment is anonymous to peers. Defaults to False.
            depth (int, optional): The depth of the comment in the thread hierarchy. Defaults to 0.
            abuse_flaggers (Optional[list[str]], optional): Users who flagged the comment. Defaults to None.
            historical_abuse_flaggers (Optional[list[str]], optional): Users historically flagged the comment.
            visible (bool, optional): Whether the comment is visible. Defaults to True.

        Returns:
            str: The ID of the inserted document.
        """
        date = datetime.now()
        comment_data = {
            "votes": self.get_votes_dict(up=[], down=[]),
            "visible": visible,
            "abuse_flaggers": abuse_flaggers or [],
            "historical_abuse_flaggers": historical_abuse_flaggers or [],
            "parent_ids": [ObjectId(parent_id)] if parent_id else [],
            "at_position_list": [],
            "body": body,
            "course_id": course_id,
            "_type": self.content_type,
            "endorsed": False,
            "anonymous": anonymous,
            "anonymous_to_peers": anonymous_to_peers,
            "author_id": author_id,
            "comment_thread_id": ObjectId(comment_thread_id),
            "child_count": 0,
            "depth": depth,
            "author_username": author_username or self.get_author_username(author_id),
            "created_at": date,
            "updated_at": date,
        }
        if parent_id:
            comment_data["parent_id"] = ObjectId(parent_id)

        comment_data["endorsement"] = None

        result = self._collection.insert_one(comment_data)

        if parent_id:
            self.update_child_count_in_parent_comment(parent_id, 1)

        self.update_comment_count_in_comment_thread(comment_thread_id, 1)

        # Notify Comment inserted
        get_handler_by_name("comment_inserted").send(
            sender=self.__class__, comment_id=str(result.inserted_id)
        )

        self.update_sk(str(result.inserted_id), parent_id)
        return str(result.inserted_id)

    def update(
        self,
        comment_id: str,
        body: Optional[str] = None,
        course_id: Optional[str] = None,
        anonymous: Optional[bool] = None,
        anonymous_to_peers: Optional[bool] = None,
        comment_thread_id: Optional[ObjectId] = None,
        at_position_list: Optional[list[str]] = None,
        visible: Optional[bool] = None,
        author_id: Optional[str] = None,
        author_username: Optional[str] = None,
        votes: Optional[dict[str, int]] = None,
        abuse_flaggers: Optional[list[str]] = None,
        historical_abuse_flaggers: Optional[list[str]] = None,
        endorsed: Optional[bool] = None,
        child_count: Optional[int] = None,
        depth: Optional[int] = None,
        closed: Optional[bool] = None,
        editing_user_id: Optional[str] = None,
        edit_reason_code: Optional[str] = None,
        endorsement_user_id: Optional[str] = None,
        sk: Optional[str] = None,
    ) -> int:
        """
        Updates a comment document in the database.

        Args:
            comment_id (ObjectId): The ID of the comment to update.
            body (Optional[str], optional): The body content of the comment.
            course_id (Optional[str], optional): The ID of the course the comment is associated with.
            anonymous (Optional[bool], optional): Whether the comment is posted anonymously.
            anonymous_to_peers (Optional[bool], optional): Whether the comment is anonymous to peers.
            comment_thread_id (Optional[ObjectId], optional): The ID of the parent comment thread.
            at_position_list (Optional[list[str]], optional): A list of positions for @mentions.
            visible (Optional[bool], optional): Whether the comment is visible.
            author_id (Optional[str], optional): The ID of the author who created the comment.
            author_username (Optional[str], optional): The username of the author.
            votes (Optional[dict[str, int]], optional): The votes for the comment.
            abuse_flaggers (Optional[list[str]], optional): A list of users who flagged the comment for abuse.
            historical_abuse_flaggers (Optional[list[str]], optional): Users who historically flagged the comment.
            endorsed (Optional[bool], optional): Whether the comment is endorsed.
            child_count (Optional[int], optional): The number of child comments.
            depth (Optional[int], optional): The depth of the comment in the thread hierarchy.

        Returns:
            int: The number of documents modified.
        """
        fields = [
            ("body", body),
            ("course_id", course_id),
            ("anonymous", anonymous),
            ("anonymous_to_peers", anonymous_to_peers),
            ("comment_thread_id", comment_thread_id),
            ("at_position_list", at_position_list),
            ("visible", visible),
            ("author_id", author_id),
            ("author_username", author_username),
            ("votes", votes),
            ("abuse_flaggers", abuse_flaggers),
            ("historical_abuse_flaggers", historical_abuse_flaggers),
            ("endorsed", endorsed),
            ("child_count", child_count),
            ("depth", depth),
            ("closed", closed),
            ("sk", sk),
        ]
        update_data: dict[str, Any] = {
            field: value for field, value in fields if value is not None
        }
        if endorsed and endorsement_user_id:
            update_data["endorsement"] = {
                "user_id": endorsement_user_id,
                "time": datetime.now(),
            }
        else:
            update_data["endorsement"] = None

        if editing_user_id:
            edit_history = []
            original_body = ""
            if comment := Comment().get(comment_id):
                edit_history = comment.get("edit_history", [])
                original_body = comment.get("body", "")
            edit_history.append(
                {
                    "author_id": editing_user_id,
                    "original_body": original_body,
                    "reason_code": edit_reason_code,
                    "editor_username": self.get_author_username(editing_user_id),
                    "created_at": datetime.now(),
                }
            )
            update_data["edit_history"] = edit_history

        update_data["updated_at"] = datetime.now()
        result = self._collection.update_one(
            {"_id": ObjectId(comment_id)},
            {"$set": update_data},
        )

        # Notify Comment updated
        get_handler_by_name("comment_updated").send(
            sender=self.__class__, comment_id=comment_id
        )

        return result.modified_count

    def delete(self, _id: str) -> int:
        """
        Deletes a comment from the database based on the id.

        Args:
            _id: The ID of the comment.

        Returns:
            The number of comments deleted.
        """
        comment = self.get(_id)
        if not comment:
            return 0

        parent_comment_id = comment.get("parent_id")
        child_comments_deleted_count = 0
        if not parent_comment_id:
            child_comments_deleted_count = self.delete_child_comments(_id)

        result = self._collection.delete_one({"_id": ObjectId(_id)})
        if parent_comment_id:
            self.update_child_count_in_parent_comment(parent_comment_id, -1)

        no_of_comments_delete = result.deleted_count + child_comments_deleted_count
        comment_thread_id = comment["comment_thread_id"]

        self.update_comment_count_in_comment_thread(
            comment_thread_id, -(int(no_of_comments_delete))
        )

        # Notify Comments deleted
        get_handler_by_name("comment_deleted").send(
            sender=self.__class__, comment_id=_id
        )

        return no_of_comments_delete

    def get_author_username(self, author_id: str) -> str | None:
        """Return username for the respective author_id(user_id)"""
        user = Users().get(author_id)
        return user.get("username") if user else None

    def delete_child_comments(self, _id: str) -> int:
        """
        Delete child comments from the database based on the id.

        Args:
            _id: The ID of the parent comment whose child comments will be deleted.

        Returns:
            The number of child comments deleted.
        """
        child_comments_to_delete = self.find({"parent_id": ObjectId(_id)})
        child_comment_ids_to_delete = [
            child_comment.get("_id") for child_comment in child_comments_to_delete
        ]
        child_comments_deleted = self._collection.delete_many(
            {"_id": {"$in": child_comment_ids_to_delete}}
        )

        for child_comment_id in child_comment_ids_to_delete:
            get_handler_by_name("comment_deleted").send(
                sender=self.__class__, comment_id=child_comment_id
            )

        return child_comments_deleted.deleted_count

    def update_child_count_in_parent_comment(self, parent_id: str, count: int) -> None:
        """
        Update(increment/decrement) child_count in parent comment.

        Args:
            parent_id: The ID of the parent comment whose child_count will be updated.
            count: It can be any number.
                   If positive, this function will increase child_count by the count.
                   If negative, this function will decrease child_count by the count.

        Returns:
            None.
        """
        update_child_count_query = {"$inc": {"child_count": count}}
        self.update_count(parent_id, update_child_count_query)

    def update_comment_count_in_comment_thread(
        self, comment_thread_id: str, count: int
    ) -> None:
        """
        Update(increment/decrement) comment_count in comment thread.

        Args:
            comment_thread_id: The ID of the comment thread
                                whose comment_count will be updated.
            count: It can be any number.
                    If positive, this function will increase comment_count by the count.
                    If negative, this function will decrease comment_count by the count.

        Returns:
            None.
        """
        update_comment_count_query = {
            "$inc": {"comment_count": count},
            "$set": {"last_activity_at": datetime.now()},
        }
        CommentThread().update_count(comment_thread_id, update_comment_count_query)

    def get_sk(self, _id: str, parent_id: Optional[str]) -> str:
        """Returns sk field."""
        if parent_id is not None:
            return f"{parent_id}-{_id}"
        return f"{_id}"

    def update_sk(self, _id: str, parent_id: Optional[str]) -> None:
        """Updates sk field."""
        sk = self.get_sk(_id, parent_id)
        self.update(_id, sk=sk)
